--- Dialplan module that uses for set up call class value to channel
--
-- @file    peerCallClass.lua
-- @author  Vyacheslav Khudan
-- @created 2015/12/07

--- @module peerCallClass
local peerCallClass = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Retrieving call class name by ident
-- @param value     A call class value (name or ident)
-- @return call class mapping value
function peerCallClass:getNameById(ident)
    local ident = tonumber(ident) 
    if not ident then
        return nil 
    end
    
    return require"common.callClass.callClassImpl":getNameById(ident)
end
--
--- Set up peer call class to channel
-- @param value     A call class name
-- @return boolean value
function peerCallClass:setUp2Channel(value)
    if not value or "" == value then
        return false
    end
    
    local callClassImpl = require"common.callClass.callClassImpl"
    
    -- process call class value
    local callClassName  = nil
    local callClassIdent = tonumber(value)
    if not callClassIdent then    -- we got call class name
        callClassName = value
        callClassIdent = callClassImpl:getIdByName(callClassName)
    else                           -- we got call class ident
        callClassName  = callClassImpl:getNameById(callClassIdent)
    end
     
    if not callClassName or not callClassIdent then
        print("Not found call class by value: %s", value)
        logger:error("Not found call class by value: %s", value)
        return false
    end
    
    -- print logs about call class data
    print("Set call class to '%s'(%s)", callClassName, callClassIdent)
    logger:debug("Set call class to '%s'(%s)", callClassName, callClassIdent)
     
    -- save call class ident to CDR structure
    channel.CDR("callclass"):set(callClassIdent)
    
    return true
end

--- Get call class value from the channel
-- @return channel call class value otherwise nil
function peerCallClass:getFromChannel()
    local rv             = nil
    local callClassIdent = channel.CDR("callclass"):get() or ""
    if "" ~= callClassIdent then
        rv = tonumber(callClassIdent)
    end
    
    return rv
end

--- Evaluate call class by normalized number
-- @param number    A number for evaluation
-- @return peer call class
function peerCallClass:evaluate(number)
    local rv = nil
    if not number or "" == number then
        return rv
    end
    
    local localCountryCode = channel.GLOBAL("GCOUNTRY_CODE"):get() or ""
    rv = require"common.callClass.callClassImpl":getClassByNumber(number, localCountryCode)
    
    return rv
end

return peerCallClass
