--- Module for WMS network calls
--
-- @file    wmsNetCall.lua
-- @author  Gennadiy Ganchev
-- @created 2017/07/13

--- @module wm
local wmsNetCall = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local argsDefines = {
     [1] = {header = "X-To-WMSNet"               , var = "pbxid"}
    ,[2] = {header = "X-Remote-Context"          , var = "ctx"}
    ,[3] = {header = "X-Remote-Calltype"         , var = "calltype"}
    ,[4] = {header = "X-Remote-Trunk"            , var = "trunkname"}
    ,[5] = {header = "X-Remote-Application"      , var = "app"}
    ,[6] = {header = "X-Remote-Application-Data" , var = "appdata"}
    ,[7] = {header = "X-Remote-Transfer"         , var = "transfer"}
    ,[8] = {header = "X-Remote-Fwd"              , var = "forward"}
    ,[9] = {header = "X-Remote-Failover-User"    , var = "failoveruser"}
    ,[10]= {header = "X-CG-Pickup-Info"          , var = "cgpickupinfo"}
    ,[11]= {header = "X-Linkedid"                , var = "linkedid"}
}
--- Preparing predial^wmsnnetwork arguments list string
-- @param  tbl   The table with arguments that should be added to string
-- @return formatted arguments list string
function wmsNetCall.makeArgs(tbl)
    if not tbl or "table" ~= type(tbl) then
        return false
    end
    local rv = {}
    for k, v in ipairs(argsDefines) do
        rv[k] = tbl[v.var] or ""
    end
    rv = table.concat(rv, ",")

    return "" ~= rv and "b(predial^wmsnetwork^1("..rv.."))"
end

function wmsNetCall.preDialRules()
    local channel = channel
    for k, v in ipairs(argsDefines) do
        local arg = channel["ARG"..k]:get() or ""
        if "pbxid" == v.var then
            arg = require"core.strConverter".fromB64(arg)
        end
        if "" ~= arg then
            channel["SIPADDHEADER_"..v.var]:set(v.header .. ": " .. arg)
        end
    end
end

return wmsNetCall