--- Module for initialization extended service strategy and excute last one
-- 
-- @file    IExServiceStrategy.lua 
-- @author  Vyacheslav Khudan
-- @created 2015/07/09

--- @module IExServiceStrategy
local IExServiceStrategy = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local strategy = {		-- meta class
	name = ""		-- Strategy name
} 

--- Check permission for Wildix Hotel call
-- @param callClass		A call class for checking
-- @return TRUE for allowed call, FALSE to deny call and NIL to make other checking
function strategy:isCallAllowed(_, callClass)
	return nil
end

--- Dialplan actions in case call was blocked
-- @param appName	A name of the called application
-- @param msg		A message printed to logs
-- @return execution of some applications
function strategy:forbidCall(_, appName, msg)
	return false
end

--- Make up room Hotel service
-- @param forbidCallback	A callback used in case service blocked
-- @param val				A value of the MUR service
-- @return execution of some applications
function strategy:serviceMUR(forbidCallback, val)
	return forbidCallback(self.name, "Not supported")
end

--- MiniBar Hotel service
-- @param forbidCallback	A callback used in case service blocked
-- @param item				A MiniBar item like article or cost
-- @param count				A count if the MiniBar items
-- @return execution of some applications
function strategy:serviceMiniBar(forbidCallback, item, count)
	return forbidCallback(self.name, "Not supported")
end

--- WakeUp Hotel service
-- @param forbidCallback	A callback used in case service blocked
-- @param mixed				A values of the WakeUp service
-- @return execution of some applications 
function strategy:serviceWakeUp(forbidCallback, ...)
	return forbidCallback(self.name, "Not supported")
end

--- Room inspection Hotel service
-- @param forbidCallback	A callback used in case service blocked
-- @param val				A value of the INSPECT service
-- @return execution of some applications
function strategy:serviceINSPECT(forbidCallback, val)
	return forbidCallback(self.name, "Not supported")
end

--- Application WakeUp implementation
-- @param forbidCallback	A callback used in case service blocked
-- @param mode 				An application mode value
-- @param arg				An application argument value
function strategy:applicationWakeUp(forbidCallback, mode, arg)
	return forbidCallback(self.name, "Not supported")
end


--- Object constructor
function IExServiceStrategy:new()
	local obj = {}
	setmetatable(obj, {__index = strategy})
	return obj
end

return IExServiceStrategy