--- Module for generating call out call file
-- 
-- @file    callOutCallFile.lua 
-- @author  Gennadiy Ganchev
-- @created 2016/07/12

--- @module callOutCallFile
local callOutCallFile = {}

--- validation of input parameters
-- @param cfg		parameters array
-- @return validation status
local function __validateParameters(cfg)
	
	local errors = 0
	local msg = ""

	if not cfg.parentName then
		errors = errors + 1
		msg = msg..",parent_name"
	end
	if not cfg.destId then
		errors = errors + 1
		msg = msg..",dest_id"
	end
	if not cfg.numId then
		errors = errors + 1
		msg = msg..",contact_numId"
	end
	if not cfg.num then
		errors = errors + 1
		msg = msg..",contact_num"
	end
	if not cfg.context then
		errors = errors + 1
		msg = msg..",context"
	end
	return errors,msg;
end


----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Generate call file
-- @param spoolDir    The result directory with call files
-- @param uniqueIdent The call out execution identifier 
-- @param contact	  A table with contact info
-- @param mode		  A call mode (real or test etc)
-- @return callFile	  A call file name
function callOutCallFile.generate(args)
    local json         = require"dkjson"
    local cf           = require"core.rawlazy.callfile".call_obj
    local base64       = require"base64"
    local rv           = false
    local spoolDir     = "/var/spool/callweaver/outgoing"
    local logger       = require"core.logger":init("[WMS 3.0 Dialplan]", 3)

    if "table" ~= type(args) then
        return rv, "No input arguments found!!"
    else
        local err,errmsg = __validateParameters(args)
        if err ~= 0 then
            return rv, "Not found some number of mandatory parameters: "..errmsg
        end
    end

    local callFileName = string.format("%s/%s-%s-%s.call", spoolDir, args.parentName, args.destId, args.numId)
    logger:debug("Generating the %s...", callFileName)

    cf:add("Channel", string.format("Local/%s:%s@wildix-callout/n", args.num, args.context))
    cf:add("Context", "wildix-callout")
    cf:add("Extension", "ivr")
    cf:add("MaxRetries", 0)
    cf:add("Set", "PARENTNAME", args.parentName)
    cf:add("Set", "DESTID", args.destId)
    cf:add("Set", "NUMID", args.numId)

    rv = cf:writefile(callFileName, true) and callFileName
    if rv then
        require"lfs".touch(callFileName)
    end
    return rv
end

return callOutCallFile

