--- Callout implementation
-- 
-- @file    callOutImpl.lua 
-- @author  P.Yadvichuk
-- @created 2016/08/30

package.path = package.path .. ";/var/lib/callweaver/dialplan/exService/callOut/?.lua"

--- @module callOutImpl
local callOutImpl = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local impl = {		        -- meta class
      mProvider     = nil
     ,mUniqueIdent  = 0
	 ,mName         = nil
     ,mColor        = nil
     ,mParams       = nil
}

--- Initialize object
function impl:start()
    self.mProvider:setProperty("status", true)
    self.mProvider:setProperty("firstDialedContact", "none")
end

--- Destroy object
function impl:destroy()
	 self.mProvider:setProperty("status", false)
    self.mProvider:delAllData()
end

-- Top object methods
function impl:ident()
    return tonumber(self.mUniqueIdent) or 0
end

-- Top object methods
function impl:name()
    return self.mName
end

-- Top object methods
function impl:color()
    return self.mColor
end

-- Top object methods
function impl:status()
    return self.mProvider:getProperty("status")
end

-- Top object methods
function impl:params(p)
    return p and self.mParams[p] or self.mParams
end

function impl:getProperty(name)
    local rv = nil
    if "uid" == name then
        rv = self:ident()
    elseif "color" == name then
        rv = self:color()
    elseif "name" == name then
        rv = self:name()
    elseif "params" == name then
        rv = self:params()
    else
        rv = self.mProvider:getProperty(name)
    end
    
    return rv
end

function impl:setProperty(name, val)
    if not name or nil == val then
        return false
    end
    
    if "uid" == name or "color" == name or "name" == name then
        return false
    end
    
    return self.mProvider:setProperty(name, val)
end

--- Methods to processing callout destination data
function impl:addDestination(data, def)
    return self.mProvider:addDestination(data, def)
end

function impl:getDestCnt()
    return self.mProvider:getDestCnt()
end

function impl:getDestIterator()
    return self.mProvider:getDestIterator()
end

function impl:getDestByNumber(number)
    return self.mProvider:getDestByNumber(number)
end

function impl:setDestParams(ident, p, v)
    return self.mProvider:setDestParams(ident, p, v)
end

function impl:getDestParams(ident, p)
    return self.mProvider:getDestParams(ident, p)
end

--- Processing contacts of the destination for callout object
function impl:getContactsCnt(destId)
    return self.mProvider:getContactsCnt(destId)
end

function impl:getContactsItertator(destId)
    return self.mProvider:getContactsItertator(destId)
end

function impl:setContactParam(destId, numId, p, v)
    return self.mProvider:setContactParam(destId, numId, p, v)
end

function impl:getContactParam(destId, numId, p)
    return self.mProvider:getContactParam(destId, numId, p)
end

function impl:setContactTry(destId, numId, v, t)
    return self.mProvider:setContactTry(destId, numId, v, t)
end

function impl:getContactTry(destId, numId, t)
    return self.mProvider:getContactTry(destId, numId, t)
end

function impl:incrContactTry(destId, numId, t)
    return self.mProvider:incrContactTry(destId, numId, t)
end

function impl:decrContactTry(destId, numId, t)
    return self.mProvider:decrContactTry(destId, numId, t)
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Constructor for callout object
-- @param nameWithColor  The object name with color property
-- @param params         The list of the object params
-- @return object with related methods otherwise nil
function callOutImpl:createNew(nameWithColor, params)
    local rv = nil
    if not nameWithColor then
        return rv
    end
    local color = nameWithColor:match("^CALLOUT%-%d+%-%d+%-(.+)$")
    if not color then
        return rv
    end
    
    local dataProvider = require"DataProvider":getInstance()
    if not dataProvider then
        return rv
    end
    
    -- initialize object
	local obj = {}
	setmetatable(obj, {__index = impl})
    
    obj.mProvider    = dataProvider
    obj.mName        = nameWithColor
    obj.mUniqueIdent = os.time()
    obj.mColor       = color
    obj.mParams      = params
    
    dataProvider:setName(nameWithColor)
    dataProvider:setProperty("uid", obj.mUniqueIdent)
    dataProvider:setProperty("color", color)
    dataProvider:setParams(params)

	return obj
end

---  Retrieve existing object
-- @param nameWithColor  The object name with color property
-- @return loaded object with related methods otherwise nil
function callOutImpl:getExisting(nameWithColor)
    local rv = nil
    if not nameWithColor then
        return rv
    end
    local color = nameWithColor:match("^CALLOUT%-%d+%-%d+%-(.+)$")
    if not color then
        return rv
    end
    
    local dataProvider = require"DataProvider":getInstance()
    if not dataProvider then
        return rv
    end
    
    local uid = dataProvider:setName(nameWithColor)
    if not uid then
        return rv
    end
    
    -- initialize object
    local obj = {}
	 setmetatable(obj, {__index = impl})
    
    obj.mProvider    = dataProvider
    obj.mName        = nameWithColor
    obj.mUniqueIdent = uid
    obj.mColor       = color
    obj.mParams      = dataProvider:getParams()
    
    return obj
end

return callOutImpl

