--- Driver to retrive astdb access via AMI 
-- 
-- @file	AMIDriver.lua 
-- @author	Vyacheslav Khudan
-- @created 2016/08/31

--- @module AMIDriver
local AMIDriver = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local function __rawRequest(o, m, k, v)
    if "table" ~= type(o) or not m or not k then
        return false
    end
    local family, key = k:match("([^/]+)/(.+)")
    return o.raw(m, {family = family, key = key, val = v})
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local driver  = require"IDriver":new()        -- meta class

--- @see description in IDriver
function driver:put(key, val)
    return __rawRequest(self, "put", key, val)
end

--- @see description in IDriver
function driver:get(key)
    return __rawRequest(self, "get", key)
end

--- @see description in IDriver
function driver:del(key)
    return __rawRequest(self, "del", key)
end

--- @see description in IDriver
function driver:delTree(key)
    return self.raw("deltree", {family = key})
end


--- Constructor
function AMIDriver:new()
	local obj = {}
	setmetatable(obj, {__index = driver})
	
	obj.raw = require"core.astman".send_dbrequest
	
	return obj
end

return AMIDriver

