--- Main callweaver route point file
-- Executed every time when core makes search extension/context and
-- to make dialplan routing
--
-- @file    extensions.lua

package.path = package.path..";/var/lib/callweaver/dialplan/?.lua"
package.path = package.path..";/var/lib/callweaver/dialplan/exService/?.lua"
package.path = package.path..";/var/lib/callweaver/dialplan/stdService/?.lua"
package.path = package.path..";/var/lib/callweaver/dialplan/stdService/?/init.lua"
package.path = package.path..";/var/lib/callweaver/dialplan/dialplanModules/?.lua"
package.path = package.path..";/var/lib/callweaver/dialplan/systemModules/?.lua"
package.path = package.path..";/var/www/provision/?.lua"

----------------------------------------------------------------
-- Channel constructor/descructor methods
----------------------------------------------------------------
-- General callback that will be executed by channel creating/destroying
function channelinit(chanData)
    -- initialize global space module
    gSpace:initialize(chanData)
    gSpace["gvars"]       = {}
    gSpace["cidAttrs"]    = {}
    gSpace["cidFeatures"] = {}
    
    -- loading modules to global space after initialization global space
    wmsldap = require"core.wmsldap"
    dLog    = require"dialplanLogger":new()
    
    if chanData then
        logger:debug(">>>>>>>>>>>> [%s] Initialize channel %s to %s@%s", CW_VERSION, channel.CHANNEL:get("name") or "", channel.EXTEN:get() or "", channel.CONTEXT:get() or "")
    else
        logger:debug(">>>>>>>>>>>> [%s] Initialization without channel ...", CW_VERSION)
    end
end

--- Callback that can be executed by channel destroy/hangup
-- Currently not executed in callweaver. You can add to hangup handler!!!
function channeldestroy()
    -- Use only syslog methods to print logs
    -- Default application NoOp does not work there!!!
    -- Must have simple functionality for debug, nothing for production logic!!!
    
    logger:debug(">>>>>>>>>>>> [%s] Destroy method in progress ...", CW_VERSION)
    logger:debug(gSpace, gSpace:getIdent())
end

----------------------------------------------------------------
-- Initialization global dialplan modules
----------------------------------------------------------------
gSpace = require"core.stateGlobalSpace":new()
logger = require"core.logger":init("[WMS 3.0 Dialplan]", 3, {requestId = "dialplan"})
ivr    = require"pbxIVR":new()

----------------------------------------------------------------
-- Other procedures
----------------------------------------------------------------
-- Table of hangup statuses 
hcause = {
     NULL           = ""    -- default cause value
    ,NOT_FOUND      = 1     -- 1 = 404 Not found
    ,NORMAL         = 16    -- 16 = Normal call clearing
    ,BUSY           = 17    -- 17 = User busy = BUSY
    ,NOT_RESPOND    = 18    -- 18 = No user responding (telephone device not connected) = NOANSWER
    ,CHANUNAVAIL    = 44    -- 19 = Temporarily unavailable = CHANUNAVAIL
    ,REJECT         = 21    -- 21 = Call rejected
    ,FAILURE        = 41    -- 41 = Temporary failure
    ,ACL_FORBIDDEN  = 85    -- TODO: ACL_FORBIDDEN is custom code. Must be added to callweaver
}
 
--- Function for load modules or use early loaded
-- @note should be remove in future
function require_once(name)
	return require(name)
end

--- Functions used for execute action after hangup
function setpostaction (taction, action)
	local postaction = nil
	if not action or "" == action then
		action = "set"
	end
	if action == "add" and gSpace["postaction"] then
		postaction = gSpace["postaction"]
		for item, value in pairs(taction) do
			postaction[item] = value
		end
		taction = postaction
		gSpace["postaction"] = taction
	elseif action == "rm" and gSpace["postaction"] then
		postaction = gSpace["postaction"]
		for _, item in pairs(taction) do
			postaction[item] = nil
		end
		taction = postaction
		gSpace["postaction"] = taction
	elseif action == "set" then 
		gSpace["postaction"] = taction
	end
end

function getpostaction ()
	return gSpace["postaction"] or {}
end

--- Dialplan logfile function
-- saved for compatibility (must be replaced by 'smartApp')
to_logfile = function(appName, ...)
    return dLog:write(appName, ...)
end

--- Internal application function with some features
function smartApp(appName, ...)
	-- XXX: must be removed after fix on vendor side
	-- Hack for w-air: answer the call to prevent no ringback after pre-answer services before dialing the other user
	local args    = {...}
	local channel = channel
	local fixArgs = nil
	
	local deviceModel = gSpace["gvars"].DeviceModel 
	if not deviceModel then
		local chanName = channel.CHANNEL("name"):get() or ""
		if not chanName:find("^Local") then
			deviceModel = (channel.SIP_HEADER("User-Agent"):get() or ""):match("Wildix%s(.-)%s.*")
		end
		if deviceModel then 
			if deviceModel:find("W-AIR") and deviceModel:find("/") then
				deviceModel = string.match(deviceModel, "(.*)/.*")
			end
		else
			deviceModel = ""
		end
		gSpace["gvars"].DeviceModel = deviceModel
	end
	
	if "Read" == appName or "Background" == appName or "Playback" == appName then
		local sound = "Read" == appName and 2 or 1
		for i = 1, #args do
			args[i] = args[i] or ""
			if "n" == args[i] or "noanswer" == args[i] then
				if deviceModel:find("W-AIR") then
					args[i] = ""
				else
					-- add silence before soundPath to avoud first word not audible
					args[sound] = (require"utils".GetSilence(2) or "") .. args[sound] or ""
				end
			end
		end
		fixArgs = table.concat(args, ",")
	end
	
	if deviceModel:find("W-AIR") and "Progress" == appName then
		appName = "Answer"
	end
	
	-- Processing dial application with extra action
	if "Dial" == appName and string.find(args[1], "SIP%/") then
	   local extraDialFlag = channel.EXTRADIALSCRIPT:get() or ""
	   if "" ~= extraDialFlag then
    	   local gvars = gSpace["gvars"]
    	   local incTrankCall = gvars.trunk_call and "" ~= gvars.trunk_call
    	   local outTrunkCall = gvars.outTrunkCall and "" ~= gvars.outTrunkCall
    	   if incTrankCall or outTrunkCall then
    	       local extraDialArg = 0
    	       if outTrunkCall then
    	           local chanName = channel.CHANNEL("name"):get() or ""
    	           extraDialArg   = chanName:match("^%w+%/(%d+)%-.*") or 1
    	       end
    	       args[3] = string.format("%sM(extradial^%s)", args[3], extraDialArg)
    	       for i = 1, #args do
    	           args[i] = args[i] or ""
	           end
    	       fixArgs = table.concat(args, ",")
    	   end
	   end
	end

	if require"core.pbxutils".in_table(appName, {"Read", "Dial", "ConfBridge", "VoiceMail"}) then
		app.CELGenUserEvent("App", appName, fixArgs or ...)
	end

	if fixArgs then
		dLog:write(appName, fixArgs)
		return app[appName](fixArgs)
	else
		dLog:write(appName, ...)
		return app[appName](...)
	end
end

--- Print message to callweaver cli
function print(...)
	local t = {}
	local msg = ""
	local args_len = 0
	for i, _ in pairs({...}) do args_len = i end
	local args_in = {...}
	for i = 1, args_len do
		if not args_in[i] then
			table.insert(t, "nil")
		elseif "table" == type(args_in[i]) then
			table.insert(t, require"core.pbxutils".tprint(args_in[i], 0, nil, true))
		elseif "string" == type(args_in[i]) then
			table.insert(t, args_in[i])
		else
			table.insert(t, tostring(args_in[i]))
		end
	end

	_, p_num = string.gsub(t[1] or "", "%%", "%%")
	if args_len > 1 and p_num > 0 then
		for i = #t, p_num  do
			table.insert(t, "nil")
		end
		msg = string.format(unpack(t))
	else
		msg = table.concat(t, " ") or arg[1]
	end

	return smartApp("NoOp", msg)
end

--- Get translation string from a global table 'gtranslation'
--- !!! Forbidden use mixed such as 'L_$lang' !!!
-- @param string A string using as a key 'gtranslation' table
-- @param mixed Mixied values to be passed to format function
-- @param lang A language used for translating (optional)
-- @return string If translation is not found return an untouched string
function _t (cstring, lang, ...)
	local channel  = channel
	local pbxutils = require"core.pbxutils"
	require"res_translation"
	
	local language
	local lang = lang or ""
	
	local use_param = lang:match("^L_[%w%-]*$") -- check case of calling function _t (with/without lang variable)
	if use_param then
		if not pbxutils.in_table(lang, {"L_en", "L_it", "L_de", "L_fr", "L_nl", "L_en-us"}) then
    		language = "en"
    	else
    		language = lang:match("L_([%w%-]*)$")
    	end
    else
		language = (channel.CHANNEL("Language"):get() or "en"):lower()
		
		if not pbxutils.in_table(language, {"en","it","de","fr", "nl", "en-us"}) then
			language = "en"
		end
    end
	
	if gtranslation[language] then
		local gt = gtranslation[language]
		if gt[cstring] then
			if use_param then
				cstring = string.format(gt[cstring], ...)
			else
				cstring = string.format(gt[cstring], lang, ...)
			end
		end
	end
	return cstring or ""
end

-- {{ Make a namespace for faxes functions
faxes = {}
function faxes:backup_for_sender(savedfax, faxpdfreport) -- Need use already with remove_control_files (sync message and action about remove files)
	if not savedfax or "" == savedfax then
		return false
	end
	
	if not require"core.diskutils".is_mount("/mnt/faxes") then
		logger:debug("Faxes device is not mounted. %s is removed", savedfax)
		return false
	end
	
	local back_path = (gSpace["cidNum"] or "") -- default fax backup path
	
	-- check global directory settings
	local fax_share_part = nil
	local fax_send_fold  = nil
	local fax_send_dir   = channel["FAX_SEND_DIR"]:get() or ""
	if "" ~=  fax_send_dir then
    	fax_share_part, fax_send_fold = fax_send_dir:match("(/mnt/faxes)/?(.*)")
    	if fax_share_part then
    		back_path = fax_send_fold or ""
    	else
    		print(string.format("FAX_SEND_DIR (%s): not included in '/mnt/faxes/'", fax_send_dir))
    	end
    end
    
    local rc       = false
    local err 	   = "Unknown status"
    local back_dir = "/mnt/faxes/" .. back_path
    local fsutils  = require"core.fsutils"
    
    -- check destiantion directory
	if not fsutils.is_dir(back_dir) then
		rc, err = fsutils.mkdir(back_dir)
		if not rc then
			local err_msg = string.format("Could not create directory %s", back_dir)
			print(msg); logger:debug("%s(%s). %s is removed.", err_msg, err, savedfax)
			return false
		end
	end
	
	-- copy fax file to destination path
	rc, err = fsutils.cpfile(savedfax, back_dir)
	if not rc then
		local err_msg = string.format("Could not copy %s to %s", savedfax, back_dir)
		print(err_msg); logger:debug("%s(%s). %s is removed.", err_msg, err, savedfax)
		return false
	end
	
	-- copy report file to destination path
	if faxpdfreport and "" ~= faxpdfreport then
		rc, err = fsutils.cpfile(faxpdfreport, back_dir)
		if not rc then
			local err_msg = string.format("Could not copy %s to %s", faxpdfreport, back_dir)
			print(err_msg); logger:debug("%s(%s). %s is removed.", err_msg, err, faxpdfreport)
		end
	end
	
	print(string.format("%s was copied to %s", savedfax, back_dir))
	
	-- update CDR field
	local lastcdrdata = faxes:getcdr()
	if lastcdrdata ~= "" then
		local filename  = savedfax:match(".*%/(.*)")
		local file_path = "F:".. back_path .. "/" ..  filename 
		faxes:setcdr(lastcdrdata:gsub("^([^|]*)||(.*)", "%1|" .. file_path .. "|%2"))
	end
	return true
end

function faxes:update_report(calluid, faxnum, faxnormnumber, faxcid, status, status_desc, remote_id, faxpages, totalpages, timestamp)
	if not calluid or not faxnum then
		return false
	end
	
	local report_file = "/tmp/mail2fax/" .. calluid .. ".csv"
	local pages = tonumber(totalpages or "") and (faxpages or "").."/"..(totalpages or "")  or ""
	
	local freport, err = io.open(report_file, "r+")
	if not freport then
		err = string.format("Can't open CSV report file %s", err)
		print(err); logger:debug(err)
		return false
	end
	
	local check_pattern = "^" .. faxnum .. ",$"
	local upd_pattern   = string.format("%s,%s,%s,%s,%s,%s,%s,%s", faxnum, faxnormnumber, faxcid, status, status_desc, remote_id, pages, timestamp)
	local tresult = {}
	for line in freport:lines() do
		if line:match(check_pattern) then
			line = line:gsub(check_pattern, upd_pattern)
		end
		table.insert(tresult, line)
	end
	
	freport:seek("set", 0) -- go to the file beginning
	local rc, err = freport:write(table.concat(tresult, "\n"))
	freport:close()
	
	if rc then
		print("Mass fax report for number", faxnum, "has been updated")
	else
		local msg = string.format("Mass fax report for number %s hasn't been updated: %s", faxnum, err)
		print(msg); logger:debug(msg)
	end
		
	return rc
end

function faxes:remove_control_files(calluid, faxpng, faxpdfreport)
	if not calluid then
		return false
	end
	
	local fsutils     = require"core.fsutils"
	local fax_dir     = "/tmp/mail2fax/0/"
	local ffile       = channel["file2fax"]:get() or ""
	local fcall       = channel["callfile"]:get() or ""
	local fcover      = channel["cover2fax"]:get() or ""
	local ftmpfaxfile = channel["tmpfaxfile"]:get() or ""
	
	-- Iterate fax directory and search linked files
	local function found_tiff_linked_files (uid, dir)
		if not uid or not dir then
			return false
		end
		
		local fsutils = require"core.fsutils"
		local file_regexp = "fax%-.*%-"..uid.."%-.*%.tiff"
		local uid_files = fsutils.readdir(dir, file_regexp, "link", true)
		if uid_files then
			return true
		end
		return false
	end
	
	-- processing control files
	if fsutils.is_link(ffile..".tiff") then	-- tiff file is link. Let's remove it and check parent tiff
		fsutils.rmfile(ffile..".tiff")
		fsutils.rmfile(faxpng)
		fsutils.rmfile(ffile..".pdf")
		
		-- check if we can remove parent tiff and png files
		local ptiff = fsutils.readdir(fax_dir, "fax%-.*%-"..calluid.."%-1%.tiff", "file", true)
		if ptiff then
			ptiff = ptiff[1]
    		local ptiff_prefix  = ptiff:match("(.*)%.tiff$"):gsub("-", "%%-")
    		local pcall_regexp  = ptiff_prefix .. "%-%d+%.call"
    		local pinuse_regexp = ptiff_prefix .. "%-%d+%.inuse"
    		
    		if not fsutils.readdir(fax_dir, {pcall_regexp, pinuse_regexp}, "file", true) and not found_tiff_linked_files(calluid, fax_dir) then
    			local faxuidname = ptiff:match("^fax%-(.*)%.tiff$")
    			local pfaxpng = "report-" .. faxuidname .. ".png"
    			local ppdf	= "fax-" .. faxuidname .. ".pdf"

    			fsutils.rmfile(fax_dir .. "/" .. ptiff)
    			fsutils.rmfile(fax_dir .. "/" .. pfaxpng)
    			fsutils.rmfile(fax_dir .. "/" .. ppdf)
    		end
    	end
	else	-- tiff file is not a link. Let's check linked files and try to remove it
		if not found_tiff_linked_files(calluid, fax_dir) then
			fsutils.rmfile(ffile..".tiff")
			fsutils.rmfile(faxpng)
			fsutils.rmfile(ffile..".pdf")
		end
	end
	fsutils.rmfile(fcall)
	fsutils.rmfile(fcover)
	fsutils.rmfile(ftmpfaxfile..".tiff")
	fsutils.rmfile(ftmpfaxfile..".pdf")
	fsutils.rmfile(faxpdfreport)
	
	print("Trying to remove " .. ffile..".tiff" .. ", " .. ffile..".pdf" .. (faxpdfreport and (" and " .. faxpdfreport) or ""))
end

-- Check parameter for attaching report or not
function faxes:report_notify_enabled(report_pdf)
	local value = channel["GFAXPDFREPORT_0"]:get()
	if "on" == value or "yes" == value then
		return report_pdf
	end
end

--- Handling CDR field with fax data
function faxes:setcdr(msg, ...)
	channel.CDR("faxdata"):set(string.format(msg, ...))
end

function faxes:getcdr(field)
	return channel.CDR(field or "faxdata"):get() or ""
end

function faxes:create_pdf_report (png_image, calluid, dstnum, remotestationid, faxstatus, faxstatusstr, faxpages, totalpages, timestamp, postfix)
	local faxreport_pdf      = nil
	local faxreport_pdf_name = nil
	if require"core.fsutils".is_readable(png_image) then
		local lang         = (channel.CHANNEL("Language"):get() or "en"):lower()
		local from_company = gSpace["gvars"].faxfromcompany or ""
		if "" == from_company then
			from_company = channel["fromcompany"]:get() or ""
		end
		local from_number, fax_header  = faxes:getStationId()
		local faxreport_rc, report_msg = require"faxreport".create_report(png_image, lang, calluid, from_number, fax_header, dstnum, from_company, remotestationid, faxstatus, faxstatusstr, faxpages, totalpages, timestamp, postfix)
		if not faxreport_rc then
			print("PDF fax report was not created: " .. report_msg or "Unkown error")
		else
			faxreport_pdf = report_msg
			faxreport_pdf_name = faxreport_pdf:match(".*%/(report%-.*%.pdf)$")
		end
	end
	return faxreport_pdf, faxreport_pdf_name
end

--- Get stationid and headerinfo from user variables
function faxes:getStationId()
	local channel = channel
	local stationid = channel["stationid"]:get() or ""
	local headerinfo = channel["headerinfo"]:get() or ""
	
	-- get fax user settings from user cache file email2fax
	local fromuid = channel["fromuid"]:get() or ""
	if fromuid ~= "" and ("" == stationid or "" == headerinfo) then
		-- processing user CTI options
		local features = require"common.userFeatures.userFeatures":new(fromuid)
        if features then
            features:load()
    		local ctiopt = features:get("CTIOPTIONS")
            if ctiopt and ctiopt ~= "" then
                local json = require"dkjson"
                local b64 = require"base64"
                ctiopt = json.decode(b64.decode(ctiopt) or "")
                if ctiopt then
                    stationid = "" == stationid and ctiopt.fax_options_stationid or stationid or ""
                    headerinfo = "" == headerinfo and ctiopt.fax_options_headerinfo or headerinfo or ""
                    gSpace["gvars"].faxfromcompany = ctiopt.fax_options_from_company or ""
                end
            end
        end
    end
	
	-- get value from channel variable of default 
	if stationid == "" then stationid = channel["fromfaxnumber"]:get() or "" end
	if stationid == "" then stationid = channel.FAX_MACHINE_ID:get() or "" end
	if stationid == "" then stationid = channel["LOCALSTATIONID0_0"]:get() or "" end
	if stationid == "" then stationid = "Wildix PBX" end
	stationid = stationid:sub(1,20)
	channel["stationid"]:set(stationid)
	if headerinfo == "" then headerinfo = channel.FAX_HEADER:get() or "" end
	if headerinfo == "" then headerinfo = channel["LOCALHEADERINFO0_0"]:get() or "" end
	if headerinfo == "" then headerinfo = "Fax Server" end
	headerinfo = headerinfo:sub(1,20)
	channel["headerinfo"]:set(headerinfo)
	
	return stationid, headerinfo
end


--- Function for getting extension attrs from LDAP
-- @param object A object by which to search in LDAP
-- @param value A string with value of search field (by default telephoneNumber)
-- @param objectclass A string with a classObject to be look up
-- @param with_dn A use all getting field (with dn)
-- @return A table with attributes from LDAP
function get_obj_attrs(object, value, objectclass, with_dn)
    if not object or "" == object then
        return {}
    end
    
    if not value then
        value = channel["LDAP_EXTEN_NUMBER"]:get() or "telephoneNumber"
	end
    
	local object_attrs, object_dn = {}, nil
	local result, err = require"core.wmsldap":get(value, object, objectclass)
    if result[1] then
    	object_dn, object_attrs = unpack(result[1])
    	return object_attrs, with_dn and object_dn
    elseif err then
    	print(err); logger:debug(err)
    	smartApp("Hangup", hcause.FAILURE)
    end
    return {}
end


local function _antitromboning(c, e)
    local dst_pbx = channel.ANTITROMBONING_SRC:get() or ""
    if dst_pbx == "" then
        print("Antitromboning mechanism has failed: destination pbx is missing")
        return false
    end

    local pbxData, errMsg = require"core.wmsldap":get("pbxDomain", dst_pbx, "wPbx")
    if errMsg then
        print("Antitromboning mechanism has failed: ", errMsg)
        return false
    end

    local dst_pbx_dn
    if type(pbxData) == "table" and type(pbxData[1]) == "table" and pbxData[1][1] then
        dst_pbx_dn = pbxData[1][1]
    end

    if not dst_pbx_dn or dst_pbx_dn == "" then
        print("Antitromboning mechanism has failed: can't get data from ldap", errMsg)
        return false
    end

    local peername = channel.CHANNEL("peername"):get() or ""
    if peername ~= "" then
        local peercontext = channel.SIPPEER(peername, "context"):get() or ""
        if peercontext ~= "" then
            channel.CHANNEL("context"):set(peercontext)
        end
    end

    channel["CALL_THROUGH_REMOTE_PBX"]:set("yes")
    channel._SIPADDHEADER_ANTITROMBONING:set("X-To-WMSNet: " .. dst_pbx_dn)
    channel.CONNECTEDLINE("name"):set(channel.ANTITROMBONING_NAME:get() or "")

    smartApp("Dial", string.format("SIP/%s@%s", e, dst_pbx))
end


local function start_monitor(recData)
    local recParams = require"dkjson".decode(require"core.strConverter".fromB64(recData))
    if "table" == type(recParams) then
        channel.MONITOR_EXEC:set(recParams["exec_prog"])
        channel.MONITOR_EXEC_ARGS:set(recParams["exec_args"])
        channel.XAD:set("true")
        channel["__IS_SYSTEM_RECORD"]:set("true")
        to_logfile("Monitor", recParams["format"], recParams["filename"], "m" .. (recParams["opts"] or ""), recParams["time_refling"], recParams["attach_format"], recParams["rec_share"])
        app.Monitor(recParams["format"], recParams["filename"], "m" .. (recParams["opts"] or ""), recParams["time_refling"], recParams["attach_format"], recParams["rec_share"])
    end
end

--------------------------------- CONTEXTS --------------------------------------
extensions = {
	["macro-autocallrec"] = {
		["s"] = function()
			require"services":start_call("autocallrec")
		end;
	};

	["macro-autocallrecpause"] = {
		["s"] = function()
			require"services":start_call("autocallrecpause")
		end;
	};

	["macro-autocallswitch"] = {
        ["s"] = function()
            require"services":start_call("autocallswitch")
        end;
    };
	
	["macro-faxcounter"] = {
		["s"] = function()
			local txfaxcnt
			local channel = channel
			local action = channel["ARG1"]:get() or ""
			if action == "++" then
				txfaxcnt = tonumber(channel.WD_TXFAXCNT:get() or "") or 0
				channel.GLOBAL("WD_TXFAXCNT"):set(txfaxcnt+1)
			elseif action == "--" then
				txfaxcnt = tonumber(channel.WD_TXFAXCNT:get() or "") or 0
				if txfaxcnt > 0 then
					channel.GLOBAL("WD_TXFAXCNT"):set(txfaxcnt-1)
				end
			else
				print("Error: invalid action for faxcounter")
			end
		end;
	};
	
    ["macro-extradial"] = {
        ["s"] = function()
            local channel  = channel
            
            -- checking remote script
            local script = channel.EXTRADIALSCRIPT:get() or ""
            if not require"core.fsutils".is_readable(script) then
                print("Not executable script:", script)
                return false
            end
            
            print("Processing extra dial script:", script)
            
            local arg1 = channel["ARG1"]:get()
            local direction = "outgoing"
            if "0" == arg1 then
                direction = "incoming"
            end
            
            -- processing users data
            local argsList = {direction = direction}
            if "incoming" == direction then
                argsList.remoteNum  = channel.CONNECTEDLINE("num"):get() or ""
                argsList.remoteName = channel.CONNECTEDLINE("name"):get() or ""
                argsList.userNum    = channel.CALLERID("num"):get() or ""
            else
                argsList.remoteNum  = channel.CALLERID("num"):get() or ""
                argsList.remoteName = channel.CALLERID("name"):get() or ""
                argsList.userNum    = arg1
            end
            
            -- retrieve data for pbx user from LDAP
            local userData  = get_obj_attrs(argsList.userNum)
            argsList.userName  = userData.cn or ""
            argsList.userEmail = userData.mail or ""
            
            --TODO: can be processed via ARG2
            argsList.event = "answer"

            argsList.callID   = channel.UNIQUEID:get() or ""
            argsList.waitTime = channel.CDR("duration"):get()
            argsList.callTags = channel.CALLTAGS:get() 
            
            -- preparing script args
            local scriptEnv = ""
            for argName, argVal in pairs(argsList) do
                scriptEnv = string.format("%s %s=\"%s\"", scriptEnv, argName:upper(), argVal)
            end
            
            local cmd = string.format("%s %s &", scriptEnv, script)
            return require"core.pbxutils".exec(cmd)
        end
	};
	
	hdlr = {
		["internal"] = function ()
			require"internalcalls":hangup()
		end;
		["service"] = function ()
			require"services":hangup()
		end;
	};

    predial = {
        ["internalcall"] = function ()
            local channel, cid_features = channel

            local user_uid = channel["ARG6"]:get() or ""
            if "" ~= user_uid then
                cid_features = require"common.userFeatures.userFeatures":new(user_uid)
                if cid_features then
                    cid_features:load()
                end
            end

            local mobility = channel["ARG1"]:get() or ""
            if mobility ~= "" then
                channel["_MOBILITY_OPTS"]:set(mobility)
            end

            require"userRingtone".preDialRules(channel["ARG2"]:get() or "")

            if cid_features then
                if cid_features:isAlternatePickupMode() then
                    local groups = cid_features:getPickupGroups()
                    local callgroups = table.concat(groups["callgroups"], ",") or ""
                    logger:debug("Setting the following callgroups:", callgroups)
                    channel.CHANNEL("namedcallgroup"):set(callgroups)
                else
                    local callgroup = channel["ARG3"]:get() or ""
                    if callgroup ~= "" then
                        channel.CHANNEL("callgroup"):set(callgroup)
                    end
                end
            end

            local trunkCall = channel["ARG4"]:get() or ""
            if trunkCall ~= "" then
                channel["trunkCall"]:set(trunkCall)
            end

            local ringOnlyActiveDevice = channel["ARG5"]:get() or ""
            if "" ~= ringOnlyActiveDevice then
                channel.SIPADDHEADER_ROAD:set("X-Ring-active-device: yes")
            end

            if cid_features then
                channel.PRESENCE_STATUS = cid_features:getUserStatus():upper()
                channel.PRESENCE_MESSAGE = require"base64".encode(
                                            require"core.pbxutils".tojson({
                                                location = tostring(cid_features:getUserStatusData("location")),
                                                message = tostring(cid_features:getUserStatusData("message")),
                                                untilTime = tostring(cid_features:getUserStatusData("untilTime"))}))
            end

            local callToLocalFailoverUser = channel["ARG7"]:get() or "false"
            if "true" == callToLocalFailoverUser then
                channel.FAILOVER_USER:set("yes")
                channel.SIPADDHEADER_LOC:set("X-IGNORE-PHONE-LOCATION: yes")
            end

            local recData = channel.DATA_FOR_RECORDING:get() or ""
            if "" ~= recData then 
                start_monitor(recData)
            end

            return extensions.predial.include()
        end;

        ["wmsnetwork"] = function ()
            require"wmsNetCall".preDialRules()
            channel["IS_CALL_THROUGH_REMOTE_PBX"]:set("true")
            local recData = channel.DATA_FOR_RECORDING:get() or ""
            if "" ~= recData then 
                start_monitor(recData)
            end
            return extensions.predial.include()
        end;

		["failover"] = function()
			local channel = channel
			local arg = channel["ARG1"]:get() or ""
			if arg == "intercom" then
				channel["VXML_URL"]:set("intercom=true")
				channel.SIPADDHEADER_CI:set("Call-Info:\ sip:wildixgw;answer-after=0")
				app.SipAddHeader("X-To-Intercom: yes")
			end
			--TODO: need improvement.
			channel.FAILOVER_USER:set("yes") -- disabling logic for ANTITRAMBONING, DIRECT_RTP as for wms user
			channel.SIPADDHEADER_LOC:set("X-IGNORE-PHONE-LOCATION: yes")

            return extensions.predial.include()
		end;
		
		["intercom"] = function ()
			local channel = channel
			channel["VXML_URL"]:set("intercom=true")
			channel.SIPADDHEADER_CI:set("Call-Info:\ sip:wildixgw;answer-after=0")
			app.SipAddHeader("X-To-Intercom: yes")
            
            return extensions.predial.include()
		end;
		
		["wmsdisa"] = function ()
			local channel = channel
			local lang =  channel["ARG1"]:get()
			channel.CHANNEL("Language"):set(lang)
            
            return extensions.predial.include()
		end;
		
		["extcall"] = function ()
			local channel = channel
			local norm_num = channel["ARG1"]:get() or ""
			if norm_num ~= "" then
				channel.CALLERID("num"):set(norm_num)
			end
			local contact_name = channel["ARG2"]:get() or ""
			if contact_name ~= "" then
			    contact_name = require"core.httpClient".urlDecode(contact_name)
				channel.CALLERID("name"):set(contact_name)
			end
			local lang = channel["ARG3"]:get() or ""
			if lang ~= "" then
				channel.CHANNEL("Language"):set(lang)
			end
            local emergencyCallFlag = channel["ARG4"]:get() or ""
            if "1" == emergencyCallFlag then
                channel.EMERGENCY_CALL:set(1)
            end
            
            local recData = channel.DATA_FOR_RECORDING:get() or ""
            if "" ~= recData then 
                start_monitor(recData)
            end

            return extensions.predial.include()
		end;
		
		["callback"] = function ()
			channel["APPLICATION"]:set("CALLBACK")
			local lang = channel["ARG1"]:get() or ""
			if lang ~= "" then
				channel.CHANNEL("Language"):set(lang)
			end
            
            return extensions.predial.include()
		end;
		
		["cloud_srv"] = function ()
			local service = channel["ARG1"]:get() or ""
			if "" ~= service then
				channel.SIPADDHEADER_CLOUD_SERV:set("X-Cloud-Service: " .. service)
			end
            
            return extensions.predial.include()
		end;
		
		["queuecalls"] = function()
		    local channel   = channel
            local encCGExts = channel.ARG1:get() or ""
            if "" ~= encCGExts then
                local json   = require"dkjson"
                local cgExts = require"base64".decode(encCGExts)
                cgExts       = json.decode(cgExts)
                if cgExts["local"] then
                    cgExts["local"] = nil
                end
                channel["_GCREMOTEEXT"]:set(json.encode(cgExts))
            end
            local cgName = channel.ARG2:get() or ""
            if "" ~= cgName then
                channel.SIPADDHEADER_CGPICKUPINFO:set("X-CG-Pickup-Info: pickup*cg_"..cgName)
                channel["_GCREMOTENAME"]:set(cgName)
            end
            
            return extensions.predial.include()
		end;

        ["parkorbit"] = function()
            channel.SIPADDHEADER_PARK:set("X-To-Routing-Point: yes")
            return extensions.predial.include()
        end;
        
        ["userringtone"] = function()
            require"userRingtone".preDialRules()
            return extensions.predial.include()
        end;
        
        include = function(c, e)    -- executing predial context defined by user (PREDIAL_CALLEE)
            local ctxName = channel.PREDIAL_INCLUDE:get() or ""
            if "" ~= ctxName then
                if not extensions[ctxName] or not extensions[ctxName]["s"] then
                    print("Invalid format of the predial include context!")
                    return false
                end
                return extensions[ctxName]["s"]()
            end
        end;
        
        --- dummy extensions (to avoiding core crash  
        -- in some cases at user predial context)
        a = function() end;
        h = function() end;
        i = function() end;
        o = function() end;
        s = function() end;
        t = function() end;
        T = function() end;
	};
	
	pbxinternal = {
		include = {"pbxfeatures", "pbxservices", "internalcalls"};
	};
	
	internalcalls = {
		["_X."] = function (c,e)
			require"services":start_call("internalcalls", e)
		end;
		["_[0-9a-zA-Z*#+]."] = function(c,e)
		    require"kite".outgoingCall(e)
		end;
	};
	
	mashup = {         -- context for incoming kite calls
	   ["_[0-9a-zA-Z*#+]."] = function(c,e)
	       require"kite".incomingCall(e)
	   end;
	};
	
	switchcalls = {    -- context for switching call after redirect
	   ["_X."] = function (c,e) 
            require"services":start_call("switchcalls", e)
        end;
	};
	
	connecttorecordcall = {    -- context for spying channel which will be recorded
	   ["_[0-9a-zA-Z]."] = function (c,e)
	       to_logfile("ChanSpy", "SIP/" .. e, "BEq")
	       app.ChanSpy("SIP/" .. e, "SBEq")
	   end;
	};
    
    anycall = {
        ["_[0-9a-zA-Z*#+]."] = function(c,e)
            app.NoCDR()
            return app.Transfer("SIP/"..e)
        end;
    };
    
    pbxsound = {
        ["_X!"] = function(c,e)
            return require"pbxSound".dialToUser(c, e)
        end;
    };
    
	mobility = {
		["_X."] = function (c,e)
		    local mobileNum       = ""
		    local annonceNum      = ""
		    local timeout         = "0"
		    local dstCntx         = ""
		    local dstLang         = "en"
		    local mobOwner        = ""
			local mobilityOptions = channel.MOBILITY_OPTS:get() or ""
			if "auto" == mobilityOptions then
                -- originate mobility call from TAPI
                local mUser     = channel.CALLERID("number"):get() or ""
                local mUserData = get_obj_attrs(mUser)
                mobileNum       = mUserData.mobile or ""
                dstCntx         = mUserData.userDialplan or ""
                mobOwner        = mUserData.owner or ""
                timeout         = "-1"
                
                -- set channel variable for correct processing
                channel.MOBILITY_OPTS:set(string.format("%s|%s|%s|%s|%s|%s", mobileNum, annonceNum, timeout, dstCntx, dstLang, mobOwner))
            else
			    mobileNum, annonceNum, timeout, dstCntx, dstLang, mobOwner = string.match(mobilityOptions , "(.*)|(.*)|(-?%d*)|(.*)|(.*)|(.*)")
		    end
            
            -- disable direct rtp if mobility confirmation is active
            if "" ~= annonceNum then
                channel["_DISABLE_DIRECTRTP"]:set(1)
            end

		    -- processing timeout
		    local constTimeOut  = "10" -- uses for freeze in case using mobile application
		    local userTimeOut   = timeout and "" ~= timeout and timeout or "0"
            if 0 <= tonumber(userTimeOut) then -- check if we have to make call freezy
                print(string.format("Added mobility call freezy in %d/%d seconds", userTimeOut, constTimeOut))
                local timeOutHeader = string.format("%s;%s;%s", e, userTimeOut, constTimeOut)
                channel["_SIPADDHEADER_MOBILITY"]:set("X-Wildix-Mobility: "..timeOutHeader)
            end
            
			-- set up associated mobility peer
			local mPeerName = e
			channel.MOBILITY_4_USER:set(mPeerName)			
			
			-- set up global variable for mobility call
			gSpace["gvars"].mobility_caller = annonceNum
			gSpace["gvars"].mobility_lang   = "" ~= dstLang and dstLang or "en"
			gSpace["gvars"].mobility_owner  = mobOwner
			gSpace["gvars"].mobility_number = mobileNum
			dstCntx 				 = "" ~= dstCntx and dstCntx or "users"
			smartApp("Goto", dstCntx, mobileNum, 1)
		end;
	};
	
	wmsdisa = {
		["_."] = function (c,e)
			local app, channel = app, channel
			channel["_APPLICATION"]:set("DISA")
			channel["_CALLTYPE"]:set("internal")
			to_logfile("Dial", "Local/" .. e .. "@" .. gSpace["gvars"].disa_context .. "/n,,THgb(predial^wmsdisa^1(" .. gSpace["gvars"].disa_lang .. "))")
			app.Dial("Local/" .. e .. "@" .. gSpace["gvars"].disa_context .. "/n,,THgb(predial^wmsdisa^1(" .. gSpace["gvars"].disa_lang .. "))")
			extensions.wmsdisa.s()
		end;
		
		["s"] = function (c,e)
			local app, channel, ivr = app, channel, ivr
			local dialstatus = channel.DIALSTATUS:get()
			if dialstatus == "BUSY" then
				ivr:tell("User busy")
			elseif dialstatus and dialstatus ~= "ANSWER" then
				ivr:tell("The dialed number is not answering")
			end
			local options = {
				max_chars = 20, 
				retry = 20,
				timeout = 5,
				failure = function () 
					return true 
				end
			}
			local num = ivr:read("vm-enter-num-to-call", options)
			if num then
				extensions.wmsdisa["_."]("wmsdisa", num)
			end
		end;
		
		["h"] = function (c,e)
			extensions.wmsdisa.s()
		end;
	};
	
	["wmsdisa-callback"] = {
		["s"] = function (c,e)
			local app, channel = app, channel
			local caller_number = channel.caller_number:get()
			local caller_context = channel.caller_context:get()
			local caller_lang = channel.caller_language:get()
			gSpace["gvars"].disa_context = caller_context
			gSpace["gvars"].disa_lang = caller_lang	--used in predial for local channel (in case preanswer service)
			to_logfile("Answer", ",nocdr")	
			app.Answer(",nocdr")
			channel["__TRANSFER_CONTEXT"]:set(caller_context)
			channel.CALLERID("num"):set(caller_number)
			channel.CHANNEL("Language"):set(caller_lang)
			return extensions.wmsdisa.s()
		end;
	};
	
	antitromboning = {	-- handling anti tromboning refer
		["_."] = _antitromboning,
		["h"] = function (c, e) return end;
	};
	
	callcompletion = { -- handling invite from remote PBX in case CC
		 ["_[+*0-9]."] = function (c, e)
		 	local channel = channel
		 	local caller_name = channel.CALLERID("name"):get() or ""
		 	channel.CALLERID("name"):set("CC, "..caller_name)
		 	smartApp("Dial", "SIP/"..e)
		 end;
	};
	
	queuecalls = {
	   ["_X!"] = function(c,e)
            require"services":start_call("queuecalls", e)
        end; 
        ["_WMS[N]et!"] = function(c,e)
            require"services":start_call("queuecalls", e)
        end;
	};
	
	["wildix-hotel"] = {	-- context used for processing wildix hotel rules (like wake-up, etc.)
		["wake-up"] = function(c, e)
			local wakeUpMode   = channel.WakeUpMode:get() or ""
			wakeUpMode		   = "" == wakeUpMode and "immediate" or wakeUpMode:gsub('"', '')
			
			local wakeUpOption = channel.WakeUpOption:get() or ""
			wakeUpOption	   = wakeUpOption:gsub('"', '')
			
			return require"services":start_call("app-wake-up", wakeUpMode, wakeUpOption);
		end 
	};
	
	["wildixtsp_internal"] = {	-- context used for processing call by tsp driver
		["_[+*0-9]!"] = function(c, e)
			local tspUser 	  = e
			local bridgedPeer = channel.BRIDGEPEER:get() or ""
			if "" ~= bridgedPeer then
				tspUser = bridgedPeer:match("SIP/(%d+)%-.*")
			end
			local tspUserAttr = get_obj_attrs(tspUser)
			if tspUserAttr.uid and tspUserAttr.userDialplan then
				local tspUserContext = extensions[tspUserAttr.userDialplan] and tspUserAttr.userDialplan or "internalcalls" 
				print(string.format("Executing wildixtsp transfer to user %q via context %q", e, tspUserContext))
				return app.Goto(tspUserContext, e, 1)
			else
				print(string.format("Invalid credentials for wildixtsp user %q!", tspUser))
				return app.Hangup(hcause.NOT_FOUND)
			end				
		end;  
	};
    
    ["wildix-callout"] = {    -- context for call out service
        ["_[+*0-9]!"] = function(c, e)
            return require"exService.callOut.callOutService".callProcess(e)
        end;
        ["ivr"] = function(c, e)
            return require"exService.callOut.callOutService".answerIVR()
        end;
        ["hHandler"] = function(c, e)
        	return require"exService.callOut.callOutService".hHandler()
        end;
    };
	
	unknowns = {
		["s"] = function() require"services":start_call("LOGIN") end;
	};
	
	["assigntouser"] = { -- context for assign user from collaboration
		["_[+*0-9]!"] = function(c, e)
			return require"stdService.login.byCollaboration".run(e, "1")
		end;
	};
	
	invalid = { -- context for invalid number
		["_."] = function(c,e)
			local text = e .. " number_is_invalid"
			ivr:tell(text, {opts = "n"})
			return app.Hangup(hcause.NOT_FOUND)
		end;
		["_[stThi]"] = function(c,e)
		end;
	};

    CustomRunner = {    -- context for processing custom application via 'Runner'
        ["s"] = function(c,e)
            local args             = channel.RUNNER_ARGS:get() or ""
            local appName, appArgs = args:match("([^%:]+)%s*%:%s*(.*)")
            if appName then
                appArgs = appArgs and require"core.pbxutils".split(appArgs, ",") or {}
                return require"services":start_call(appName, unpack(appArgs))
            end
        end;
    };
	
	outgoingfax = {
		["_[+*0-9]!"] = function (c,e)
			local app, channel = app, channel
			local pbxutils = require"core.pbxutils"
			local fsutils = require"core.fsutils"
			
			if e == nil then
				print("SORRY. NO FAXNUMBER. HANGUP")
				to_logfile("Hangup", hcause.REJECT)
				app.Hangup(hcause.REJECT)
			end
			
			channel["__LDAP_OFFICE_NUMBER"]:set('facsimileTelephoneNumber')
			channel["__APPLICATION"]:set("FAX")
			
			local fromnumber = channel["fromnumber"]:get()
			if fromnumber then
				channel.CALLERID("num"):set(fromnumber)
				gSpace["cidNum"] = fromnumber
			end
			
			-- Increase global counter of sending faxes
			app.MacroExclusive("faxcounter", "++")
			
			local callfile = channel["callfile"]:get() or ""
			local callname = callfile:match("(.*)\.call$")
			if fsutils.is_readable(callfile) then
				callfile = callname .. ".inuse"
				fsutils.mvfile(callname .. ".call", callfile)
				channel["callfile"]:set(callfile)	-- save for correct remove_control_files
			end
			local faxnumberid = callname:gsub("-", ""):match("%/(fax.*)$") or ""
			channel["_faxnumberid"]:set(faxnumberid) -- create faxnumberid and push it for using normalizaton of faxnumber and return dialstatus
			local dst_context = channel["context"]:get() or "users"
			
			to_logfile("Dial", "Local/"..e.."@"..dst_context)
			app.Dial("Local/"..e.."@"..dst_context)
			
			-- detecting dialstatus for FAX call
			local dialstatus, hcausecode  -- parce this value from global variable
			local fax_gvar_stat = channel[faxnumberid.."_stat"]:get() or ""
			if fax_gvar_stat ~= "" then
				dialstatus, hcausecode = fax_gvar_stat:match("(.*)|(.*)")
			end
			if not dialstatus then
				dialstatus = channel["DIALSTATUS"]:get() or ""
			else
				channel["DIALSTATUS"]:set(dialstatus) -- set dialstatus getting from child channel
			end
			print("DIALSTATUS is", dialstatus)
			
			-- processing FAX call in case remote side didn't answer
			if dialstatus ~= "ANSWER" then
				local faxnormnumber, faxcontact_name, src_name -- parse this value from global variable
				local fax_gvar_pb = channel[faxnumberid.."_pb"]:get() or ""
				if fax_gvar_pb ~= "" then
					faxnormnumber, faxcontact_name, src_name = fax_gvar_pb:match("(.*)|(.*)|(.*)")
				end
				if not faxnormnumber or faxnormnumber == "NULL" then
					faxnormnumber = e
				end
				
				local com_id 		  = 0
				local file2fax 		  = channel["file2fax"]:get() or ""
				local faxdir, faxuid  = file2fax:match("(.*%/)fax%-(.*)$")
				local faxpng		  = faxdir.."report-"..faxuid..".png"
    			
    			local calluid, trycount = callname:match(".*%-(%d+)%-%d+%-(%d+)")
    			trycount = tonumber(trycount) or 0
    			local is_report = fsutils.is_readable("/tmp/mail2fax/" .. calluid .. ".csv")
                
                -- set correct CDR data
				channel.CDR("c_to"):set(faxnormnumber or "")
				channel.CDR("c_from"):set(fromnumber or "")
				channel.CDR("to_name"):set(faxcontact_name or "")
				channel.CDR("from_name"):set(src_name or "")
				faxes:setcdr("ERR||%s|%s", e, dialstatus)  -- will be updated by 'backup_for_sender'
    			
    			local sender = ""
    			if not is_report then -- checking sender mails
    				sender = channel["sender"]:get() or ""
        			local bksender = channel["bkmailbox"]:get() or ""
        			local bkmailbox = channel["GBKMAILBOX_0"]:get() or ""	-- check backup mailbox GBKMAILBOX_0 (set in the wms fax/server)
        			if "" ~= bkmailbox then
        				sender = sender.." "..bkmailbox
        			end
        			if "" ~= bksender then
        				sender = sender.." "..bksender
        			end
    			end
				
				-- handle fax transmissions next step for CHANUNAVAIL dialstatus or Max attempts reached
				local need_notification = false
				if 0 >= trycount then
					need_notification = "MAXATTEMPTS"
				elseif dialstatus == "CHANUNAVAIL" then
					local ext_stat = channel.EXTSTATUS:get() or ""
					if ext_stat == "NOTFOUND" or hcausecode == "1" or hcausecode == "1D" then
						need_notification = "NOTFOUND"
					end
				end
                
				if need_notification then -- sending notification about error
                    local timestamp     = os.date("%Y-%m-%d %X")
					local totalpages    = channel["totalpages"]:get() or ""
                    local faxreport_pdf = nil
					
					-- set advanced faxstatus
					local faxstatus_adv = ""
					if "NOTFOUND" == need_notification then
						faxstatus_adv = hcausecode == "1" and "Unallocated number" or "Number not present in dialplan"
					else
						faxstatus_adv = string.format("%s(Max number of attempts reached)", dialstatus)
					end
					
					if "NOTFOUND" == need_notification then
						if hcausecode == "1" then
							print("Unallocated number ("..e.."@"..dst_context..").")
						else
							print("Number ("..e.."@"..dst_context..") not present in dialplan.")
						end
					else
						print("Max number of attempts reached...")
					end
                    
                    --convert fax file to pdf for backup and email notification
                    pbxutils.run("chrt -o 0 wms_attach_converter %q %s %s 2>/dev/null", file2fax..".tiff", "pdf", "keepSrc")
                    
					if is_report then -- update CSV report
						faxes:update_report(calluid, e, faxnormnumber, faxcontact_name or "", "FAILED", faxstatus_adv, "", 0, totalpages, timestamp)
					else
						channel.CHANNEL("Language"):set(channel["faxlanguage"]:get() or "en")
						-- try create PDF report
    					faxreport_pdf = faxes:create_pdf_report(faxpng, calluid, faxnormnumber, "", "FAILED", faxstatus_adv, 0, totalpages, timestamp)
    					if sender ~= "" then  -- try create and send mail notification
    						local msubj = _t("msubj.FAXSENDFAIL", faxnormnumber)
    						local mbody = ""
    						local mail_report = faxes:report_notify_enabled(faxreport_pdf)
    						if not mail_report then
    							if "NOTFOUND" == need_notification then
    								mbody = _t(hcausecode == "1" and "mbody.FAXSTATUSFAILT" or "mbody.FAXSTATUSFAILD", timestamp, faxnormnumber)
    							else
    								mbody = _t("mbody.FAXSTATUSFAIL", timestamp, faxnormnumber, dialstatus) .. _t("string.TRIESREACHED")
    							end
    						end
							pbxutils.mailto(mbody, msubj, sender, {file2fax..".pdf", mail_report}, com_id)
						end
					end
                    
                    faxes:backup_for_sender(file2fax..".pdf", faxreport_pdf)
    				faxes:remove_control_files(calluid, faxpng, faxreport_pdf)
                elseif fsutils.is_readable(callfile) then   -- updating call file with fax data
					-- Les't update callfile data
					local cf = require"core.rawlazy.callfile".call_obj
					local rc, err = cf:readfile(callfile)
					if rc then
						rc, err = cf:upd("Set", "prevcallstatus", dialstatus)
						if err then logger:debug("Callfile update: %s", err) end
						
						rc, err = cf:upd("Set", "faxnormnumber", faxnormnumber)
						if err then logger:debug("Callfile update: %s", err) end
						
						rc, err = cf:writefile(callfile)
						if err then logger:debug("Callfile write to file: %s", err) end
					elseif err then
						logger:debug(err, "Callfile read from file")
					end
					
					-- Set call name for new file
					local prefix, trycount = callfile:match("(.*)-(.*)\.inuse$")
					trycount = tonumber(trycount) or 0
					local newcallname = prefix.."-"..trycount
					local congestionTry = tonumber(channel["GFAXCONGESTION_TRY"]:get()) or 30
					local noansretry = tonumber(channel["noansretry"]:get()) or 5
					if hcausecode == "34" and channel["prevcallstatus"]:get() ~= "CONGESTION" then	-- restore fax retry used for check trunk call limitation
						newcallname = prefix .."-".. congestionTry
					elseif hcausecode ~= "34" and
					    channel["prevcallstatus"]:get() == "CONGESTION" and
					    dialstatus ~= "CONGESTION" and
					    trycount > noansretry then -- restore fax retry back to noansretry value
					    newcallname = prefix .."-".. noansretry
					elseif dialstatus == "CONGESTION" and channel["prevcallstatus"]:get() ~= dialstatus then
						newcallname = prefix .."-"..tostring(noansretry - 1)	-- up to NOANSRETRY tries
					end
					fsutils.mvfile(callfile, newcallname .. ".call")
				end
                
                -- clear global variables
                channel.GLOBAL(faxnumberid.."_pb"):set("NULL")
				channel.GLOBAL(faxnumberid.."_stat"):set("NULL")
                
				-- Decrease global counter of sending faxes
				app.MacroExclusive("faxcounter", "--")
			end
		end;
		
		["h"] = function (c,e)	
			local dialstatus = channel.DIALSTATUS:get() or ""
			if dialstatus == "ANSWER" then	-- push CDR only if not answered
				to_logfile("NoCDR", "")
				app.NoCDR()
			end
		end;
	};
	
	faxout = {
		["_[+*0-9]!"] = function(c,e)
			local app, channel = app, channel
			local pbxutils = require"core.pbxutils"
			local fsutils = require"core.fsutils"
			
			-- we must get src and dst number form channel variables because original move
			local fromnumber = channel["fromnumber"]:get() or ""
			if fromnumber ~= "" then
				channel.CALLERID("num"):set(fromnumber)
				gSpace["cidNum"] = fromnumber
			end
			
			 -- getting faxnumberid, normalization faxnumber and remove global variables
			local callfile = channel["callfile"]:get() or ""
			local callname = callfile:match("(.*)\.call$") or ""
			gSpace["gvars"].faxcallname = callname
			local faxnumberid = callname:gsub("-", ""):match("%/(fax.*)$") or ""
			local fax_gvar_pb = channel[faxnumberid.."_pb"]:get() or ""
			channel.GLOBAL(faxnumberid.."_pb"):set("NULL")
			local faxnormnumber
			if fax_gvar_pb ~= "" then
				faxnormnumber, gSpace["gvars"].faxcontact_name, gSpace["gvars"].faxsrc_name = fax_gvar_pb:match("(.*)|(.*)|(.*)")
			end
			if not faxnormnumber or faxnormnumber == "NULL" then
				faxnormnumber = e
			end
			gSpace["gvars"].faxnormnumber = faxnormnumber
			gSpace["gvars"].faxnum        = e	-- used for fax mass report
			
			local stationid, headerinfo = faxes:getStationId()
			local fromuid = channel["fromuid"]:get() or ""
			if fromuid ~= "" then
				-- let's write getting value of the stationid and headerinfo to callfile
				local cf = require"core.rawlazy.callfile".call_obj
				local rc, err = cf:readfile(callname..".inuse")
				if rc then
					rc, err = cf:del("Set", "fromuid")
					if err then logger:debug("Callfile delete: %s", err) end
					
					rc, err = cf:upd("Set", "stationid", stationid)
					if err then logger:debug("Callfile update: %s", err) end
					
					rc, err = cf:upd("Set", "headerinfo", headerinfo)
					if err then logger:debug("Callfile update: %s", err) end
					
					rc, err = cf:upd("Set", "fromcompany", gSpace["gvars"].faxfromcompany)
					if err then logger:debug("Callfile update: %s", err) end
					
					rc, err = cf:writefile(callname..".inuse")
					if err then logger:debug("Callfile write to file: %s", err) end
				elseif err then 
					logger:debug(err, "Callfile read from file")
				end
			end
			channel.FAXOPT("localstationid"):set(stationid)
			channel.FAXOPT("headerinfo"):set(headerinfo)
			
			print("start faxing to", faxnormnumber)
			gSpace["gvars"].faxtime = os.time()
			
			local file2fax 		 = channel["file2fax"]:get() or ""
			local faxdir, faxuid = file2fax:match("(.*%/)(.*)$")
			
			gSpace["gvars"].file2fax = file2fax
			gSpace["gvars"].faxdir 	 = faxdir
			
			local cover2fax 	 = channel["cover2fax"]:get() or ""
			local tmpfaxfile  	 = file2fax
			if cover2fax ~= "" then
				tmpfaxfile = "/tmp/" .. faxuid
				pbxutils.exec("tiffcp %s %s %s", cover2fax, file2fax..".tiff", tmpfaxfile..".tiff")
			end
			
			gSpace["gvars"].tmpfaxfile = tmpfaxfile
			
			-- set normalize number and contact name getting from fax global var
			channel.CONNECTEDLINE("all"):set(string.format([===["%s" <%s>]===], gSpace["gvars"].faxcontact_name or "", faxnormnumber or ""))
			
			-- Enable debug (use with enable option fax in logger.conf)
			local faxdebug = channel.FAXDEBUG:get() or ""
			faxdebug = faxdebug == "yes" and "d" or ""
			
			-- Decrease global counter of sending faxes 
			-- and move variable control to callweaver
			app.MacroExclusive("faxcounter", "--")
			
			to_logfile("SendFAX", tmpfaxfile..".tiff", "fzs"..faxdebug)
			app.SendFAX(tmpfaxfile..".tiff", "fzs"..faxdebug)
		end;
			
		["h"] = function (c,e)
			local app, channel = app, channel
			local pbxutils = require"core.pbxutils"
			local fsutils = require"core.fsutils"

			local com_id = 0
			local file2fax = gSpace["gvars"].file2fax
			local tmpfaxfile = gSpace["gvars"].tmpfaxfile
			local totalpages = channel["totalpages"]:get() or ""
			local dstnum = gSpace["gvars"].faxnormnumber or ""
			local dstname = gSpace["gvars"].faxcontact_name or ""
			
			print("stop faxing", dstnum, tmpfaxfile..".tiff")
			
			if tmpfaxfile and tmpfaxfile ~= file2fax then
				channel["tmpfaxfile"]:set(tmpfaxfile)	-- save it for correct backup to user share
				file2fax = tmpfaxfile					-- as fax attachment user combined fax file
			end
			
			--convert sent fax file (merged with cover) to pdf for backup 
			pbxutils.run("chrt -o 0 wms_attach_converter %q %s %s 2>/dev/null", tmpfaxfile..".tiff", "pdf", "keepSrc")
			
			-- create new CDR struct (with saving duration time, etc.)
			to_logfile("ResetCDR", "t")
			app.ResetCDR("t")
			
			-- set correct CDR fields
			channel.CDR("c_to"):set(dstnum)
			channel.CDR("to_name"):set(dstname)
			channel.CDR("c_from"):set(gSpace["cidNum"] or "")
			channel.CDR("from_name"):set(gSpace["gvars"].faxsrc_name or "")
			channel.CDR("pbxid"):set(pbxutils:pbxSerial())
			
			channel.CHANNEL("Language"):set(channel["faxlanguage"]:get() or "en")
			
			dstname = dstname ~= "" and '"'..dstname..'"' or "" -- update format for notification
			local faxdir, faxname = gSpace["gvars"].file2fax:match("(.*\/)(.*)$") 
			local faxstatus = channel.FAXOPT("status"):get() or ""
			local faxstatusstr = channel.FAXOPT("statusstr"):get() or ""
			local faxpages = tonumber(channel.FAXOPT("pages"):get() or "") or 0
			local remotestationid = channel.FAXOPT("remotestationid"):get() or ""
			local faxpng = (faxname or ""):match("^fax%-(.*)$")
			if faxpng then
				faxpng = gSpace["gvars"].faxdir .. "report-"..faxpng..".png"
			end
			
			local callname = gSpace["gvars"].faxcallname
			local callfile = callname..".inuse" 		-- update name (for faxout we already dial)
			channel["callfile"]:set(callfile)	-- save for correct remove_control_files
			
			local timestamp = os.date("%Y-%m-%d %X")
			local calluid = callname:match(".*%-(%d+)%-%d+%-%d+") or ""
			local is_report = fsutils.is_readable("/tmp/mail2fax/" .. calluid .. ".csv")
			
			local sender = ""
			if not is_report then	-- checking sender mails
    			sender = channel["sender"]:get() or ""
    			local bksender = channel["bkmailbox"]:get() or ""
    			local bkmailbox = channel["GBKMAILBOX_0"]:get() or ""	-- check backup mailbox GBKMAILBOX_0 (set in the wms fax/server)
    			if "" ~= bkmailbox then
    				sender = sender.." "..bkmailbox
    			end
    			if "" ~= bksender then
    				sender = sender.." "..bksender
    			end
    		end
			
			local msubj = ""
			local mbody = ""
			local faxreport_pdf      = nil
			local faxreport_pdf_name = nil
			
			-- process fax delivery notification
			if faxstatus == "FAILED" and faxstatusstr == "TIFF/F file cannot be opened" then
				print("Fax sending was not successful to number", dstnum, "Note:", faxstatusstr)
				
				if is_report then
					faxes:update_report(calluid, gSpace["gvars"].faxnum, dstnum, gSpace["gvars"].faxcontact_name or "", faxstatus, faxstatusstr, remotestationid, faxpages, totalpages, timestamp)
				else
					-- try create PDF report
					faxreport_pdf, faxreport_pdf_name = faxes:create_pdf_report(faxpng, calluid, dstnum, remotestationid, faxstatus, faxstatusstr, faxpages, totalpages, timestamp)
					if sender ~= "" then
    					msubj = _t("msubj.FAXNOTSENT", dstname, dstnum)
    					local mail_report = faxes:report_notify_enabled(faxreport_pdf)
    					if not mail_report then
    						mbody = _t("mbody.FAXSTATUS", timestamp, remotestationid, dstname, dstnum, faxpages, totalpages, faxstatusstr)
    					end
    					pbxutils.mailto(mbody, msubj, sender, {file2fax..".pdf", mail_report}, com_id)
    				end
				end
				
				-- set CDR data
				faxes:setcdr("ERR||%s|%s", dstnum, faxstatusstr)
				
				-- try save report or faxfile and remove control files
				faxes:backup_for_sender(file2fax..".pdf", faxreport_pdf)				
				faxes:remove_control_files(calluid, faxpng, faxreport_pdf)
			elseif faxstatus ~= "SUCCESS" then
				local trycount = tonumber(callfile:match(".*%-(.*)\.inuse$")) or 0
				
				-- set advanced fax status
				local faxstatus_adv = faxstatusstr
				if 0 == faxpages then
					if 0 >= trycount then -- it was last attempt
						faxstatus_adv = string.format("%s(%s)", faxstatus_adv, "Max number of attempts reached")
					end
				else
					faxstatus_adv = string.format("%s(%s)", faxstatus_adv, "Outgoing fax partially delivered")
				end
				
				print("Fax sending was not successful to number", dstnum, "Note:", faxstatus_adv)
				
				if is_report then	       -- try update CSV report
					faxes:update_report(calluid, gSpace["gvars"].faxnum, dstnum, gSpace["gvars"].faxcontact_name or "", faxstatus, faxstatus_adv, remotestationid, faxpages, totalpages, timestamp)
				else
					-- try create PDF report
					faxreport_pdf, faxreport_pdf_name = faxes:create_pdf_report(faxpng, calluid, dstnum, remotestationid, faxstatus, faxstatus_adv, faxpages, totalpages, timestamp)
					local mail_report = faxes:report_notify_enabled(faxreport_pdf)
					if sender ~= "" then  -- try create and send mail notification 
    					if not mail_report then
    						mbody = is_report and "" or _t("mbody.FAXSTATUS", timestamp, remotestationid, dstname, dstnum, faxpages, totalpages, faxstatusstr)
    					end
    					if faxpages == 0 then
    						msubj = is_report and "" or _t("msubj.FAXNOTSENT", dstname, dstnum)
    						if not mail_report and faxstatus_adv:find("Max number of attempts reached") then
    							mbody = mbody .. _t("string.TRIESREACHED")
        					end
    					else
    						msubj = is_report and "" or _t("msubj.FAXSENTPARTIAL", dstname, dstnum)
    						if not mail_report and faxstatus_adv:find("Outgoing fax partially delivered") then
    							mbody = mbody .. _t("string.PARTIALSEND")
    						end
    					end
    					pbxutils.mailto(mbody, msubj, sender, {file2fax..".pdf", mail_report}, com_id)
    				end
				end
				
				-- try update callfile by last callstatus
				if faxpages == 0 and 0 < trycount and fsutils.is_readable(callfile) then
					local cf = require"core.rawlazy.callfile".call_obj
					local rc, err = cf:readfile(callfile)
					if rc then
    					rc, err = cf:upd("Set", "prevcallstatus", "ANSWER")
    					if err then logger:debug("Callfile update: %s", err) end
    					
    					rc, err = cf:writefile(callfile)
    					if err then logger:debug("Callfile write to file: %s", err) end
    				elseif err then 
    					logger:debug(err, "Callfile read from file")
    				end
    				local newcallname = callname 
					if channel["prevcallstatus"]:get() ~= "ANSWER" then
						local ansretry    = channel["ansretry"]:get() or ""
						local prefix      = callfile:match("(.*)%-.*%.inuse$")
						newcallname = prefix .. "-" .. tostring(tonumber(ansretry) - 1) -- up to ANSRETRY tries
					end
					fsutils.mvfile(callfile, newcallname .. ".call")
				end
				
				-- set CDR data
				faxes:setcdr("ERR||%s|%s", dstnum, faxstatusstr)
				
				-- try save report and remove control files
				faxes:backup_for_sender(file2fax..".pdf", faxreport_pdf)
				if faxpages ~= 0 or 0 >= trycount then -- partial sent, remove all control files
					faxes:remove_control_files(calluid, faxpng, faxreport_pdf)
				else	-- remove some files, like PDF report and merged fax file
					if faxreport_pdf then
						fsutils.rmfile(faxreport_pdf)
						print("Remove", faxreport_pdf)
					end
					local tmpfaxfile = channel["tmpfaxfile"]:get() or ""
					if "" ~= tmpfaxfile then
						fsutils.rmfile(tmpfaxfile..".tiff")
						fsutils.rmfile(tmpfaxfile..".pdf")
					end
				end
			else -- handling success transmission
				local duration = os.difftime(os.time(), gSpace["gvars"].faxtime)
				local faxresolution = channel.FAXOPT("resolution"):get() or ""
				local faxpages_bad = tonumber(channel.FAXOPT("bad_pages"):get() or "") or 0
				
				-- set advanced fax status
				local faxstatus_adv = faxstatusstr
				if faxpages_bad > 0 then
					faxstatus_adv = string.format("%s with %d error pages", faxstatus_adv, faxpages_bad)
				end

				if faxpages == 0 then -- set status to "FAILED" in case 0 pages were sent.
                    msubj         = _t("msubj.FAXSENDFAIL", dstname, dstnum)
                    faxstatus     = "FAILED"
                    faxstatus_adv = "FAILED"
                    faxstatusstr  = "FAILED"
				end
				print("Fax transmission was %s successful to number", (0==faxpages and "not" or ""), dstnum, "Note:", faxstatus_adv)
				
				if is_report then        -- try update CSV report
					faxes:update_report(calluid, gSpace["gvars"].faxnum, dstnum, gSpace["gvars"].faxcontact_name or "", faxstatus, faxstatus_adv, remotestationid, faxpages, totalpages, timestamp)
				else
					-- try create PDF report
					faxreport_pdf, faxreport_pdf_name = faxes:create_pdf_report(faxpng, calluid, dstnum, remotestationid, faxstatus, faxstatus_adv, faxpages, totalpages, timestamp)
					if sender ~= "" then -- try create and send mail notification 
    					if "" == msubj then 
    					   msubj = _t("msubj.FAXSENTOK", dstname, dstnum)
    					end
    					local mail_report = faxes:report_notify_enabled(faxreport_pdf)
    					if not mail_report then
    						if faxpages_bad > 0 then
    							mbody = _t("mbody.FAXSTATUSERR", timestamp, remotestationid, dstname, dstnum, faxpages, totalpages, faxpages_bad, faxstatusstr)
    						else
    							mbody = _t("mbody.FAXSTATUS", timestamp, remotestationid, dstname, dstnum, faxpages, totalpages, faxstatusstr)
    						end
    					end
    					pbxutils.mailto(mbody, msubj, sender, {file2fax..".pdf", mail_report}, com_id)
    				end
				end
				
				-- set CDR data
				if faxpages == 0 then
				    faxes:setcdr("ERR||%s|%s", dstnum, faxstatusstr)
				else
				    faxes:setcdr("OK||%s|%s|%s|%s|%s|%s", dstnum, faxpages, faxresolution, faxstatusstr, remotestationid, duration)
				end
				
				-- try save report or faxfile and remove control files
				faxes:backup_for_sender(file2fax..".pdf", faxreport_pdf)
				faxes:remove_control_files(calluid, faxpng, faxreport_pdf)
			end
		end;
	};

	["disable-autoanswer"] = { -- context for paging to users with active call (WMS-5818)
		["_[+0-9]."] = function(c, e)
			local force = nil
			local i = 1
			while i < 20 do
				var_name = "SIPADDHEADER" .. ( #tostring(i) == 2 and tostring(i) or "0" .. tostring(i) )
				sipheader = channel[var_name]:get()
				if sipheader and "" ~= sipheader then 
					if sipheader:find("answer%-after") and sipheader:find(";force") then
						force = true
						print("A forced auto-answer has been detected.")
						break
					end
				end
				i = i + 1
			end
			if not force then
				app.SipRemoveHeader("Call-Info")
			end
			return app.Dial("SIP/" .. e)
		end;
	};

	["classound-internal"] = {
		["_[+0-9]."] = function(c, e)
			local callid = channel.SIP_HEADER("Call-ID"):get() or ""
			if "" ~= callid then
				channel["__X-Forward"]:set(callid)
			else
				logger:warning("Can not get call-id for the call from classound")
			end
			return app.Goto("classound", e, 1)
		end;
	};

	["returnTransferCtx"] = {
		["_[+0-9]."] = function (c, e)
			local transferNum = channel["TRANSFER_NUMBER"]:get() or ""
			transferNum = transferNum ~= "" and transferNum
			local dst = channel.CONNECTEDLINE("num"):get() or ""
			local dst_attrs = get_obj_attrs(dst)
			local retMode = dst_attrs.telephoneNumber and "CALL" or "NONUSER"
			local rftCtx  = dst_attrs.telephoneNumber and "localUser" or "failExten"
			return require"internalcalls":return_from_transfer(transferNum, dst, retMode, rftCtx)
		end;
	};
}

--- Loading additional extensions (WMS modified)
require"dialplanLoader":proceed("extensions-incomingcalls") 
require"dialplanLoader":proceed("res_features")

-- update unknowns context for special configuration requirements
if "table" == type(extensions["login-emergency"]) then

    -- remove included internalcalls context
    local function _remove_internal(include)
        local rv = {}
        for num, val in pairs(include) do
            if "internalcalls" ~= val and "pbxinternal" ~= val then
                rv[num] = val
            end
        end
        return rv
    end

    if "table" == type(extensions["login-emergency"]["include"]) then
        extensions["login-emergency"]["include"] = _remove_internal(extensions["login-emergency"]["include"])
    end

    -- import configuration from user configured context
    for ext, data in pairs(extensions["login-emergency"]) do
        if not extensions["unknowns"][ext] then
            extensions["unknowns"][ext] = data
        end
    end
    extensions["login-emergency"] = nil
    
    -- add pbxservices login extension
    local loginCode = gl_features.LOGIN and gl_features.LOGIN.access
    if loginCode then
        extensions["unknowns"][loginCode]            = extensions["unknowns"]["s"]
        extensions["unknowns"]["_"..loginCode.."!"]  = extensions["unknowns"]["s"]
        extensions["unknowns"]["_"..loginCode.."*!"] = extensions["unknowns"]["s"]
    end
end

-- static functions for pbxservices/pbxfeatures
if type(extensions["pbxservices"]) == "table" then
    extensions["pbxservices"]["T"] = function()
    	if gSpace["gvars"].application == "KITE" then
    		smartApp("Playback", "beep")
    		smartApp("Hangup", hcause.NORMAL)
    	end
    end
    extensions["pbxservices"]["_ri[n]gto[n]e!"] = function(c,e)
        return require"userRingtone".play(c, e)
    end
    extensions["pbxservices"]["_vmplay."] = function(c,e)
        local vmApp	= require"common.voiceMailApplication"
        local vmbox, vmtype, vmfile = (e or ""):match("vmplay(%d+)(%a%a%a)(.*)")
        local cid = gSpace["cidNum"] or channel.CALLERID("num"):get()

        if not (vmbox and vmtype and vmfile) then
            print("Wrong vmplay request:", e or "")
            logger:error("Wrong vmplay request: %s", e or "")
            return smartApp("Hangup", hcause.FAILURE)
        end

        local cid_attrs = get_obj_attrs(cid)
        if cid_attrs.telephoneNumber == nil then
            print("Caller user is not found in ldap")
            logger:warning("Caller user is not found in ldap")
            return smartApp("Hangup", hcause.NOT_FOUND)
        end
        local cid_features = require"common.userFeatures.userFeatures":new(cid_attrs.uid)
        if not cid_features then
            print("Error loading features of the caller user with uid %s", cid_attrs.uid)
            logger:error("Error loading features of the caller user with uid %s", cid_attrs.uid)
            smartApp("Hangup", hcause.NOT_FOUND)
        end
        cid_features:load()

        local foreign_vm
        if "userFeatures" == cid_features.type and cid_features:get("SHAREDVM") then
            local svm = cid_features:get("SHAREDVM")
            if vmbox == svm then
                cid_attrs = get_obj_attrs(svm)
                if cid_attrs.telephoneNumber == nil then
                    print("Caller user is not found in ldap")
                    logger:warning("Caller user is not found in ldap")
                    return smartApp("Hangup", hcause.NOT_FOUND)
                end
            else
                foreign_vm = true
            end
        else
            if vmbox ~= cid then
                foreign_vm = true
            end
        end

        if "yes" ~= require"common.permissions":check(cid_attrs["owner"] or "", "canuse", "VOICEMAIL", nil, cid_attrs["licenseType"]) then
            print("Voicemail rejected by group permissions (cannot use Voicemail)")
            smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end
        if foreign_vm then
            local dst_attrs = get_obj_attrs(vmbox)
            if dst_attrs.telephoneNumber == nil then
                print("Called user is not found in ldap")
                logger:warning("Called user is not found in ldap")
                return smartApp("Hangup", hcause.FAILURE)
            end
            if "yes" ~= require"common.permissions":check(cid_attrs["owner"] or "", "yesno", "SEE", dst_attrs["owner"], cid_attrs["licenseType"]) then
                print("Voicemail rejected by group permissions (cannot View)")
                return smartApp("Hangup", hcause.ACL_FORBIDDEN)
            end
            if "yes" ~= require"common.permissions":check(cid_attrs["owner"] or "", "yesno", "SEE_VOICEMAIL", dst_attrs["owner"], cid_attrs["licenseType"]) then
                print("Voicemail rejected by group permissions (cannot See voicemail)")
                return smartApp("Hangup", hcause.ACL_FORBIDDEN)
            end
        end

        app.NoCDR()

        local vm_path = "/var/spool/callweaver/voicemail/" .. vmbox .. "/Old/"
        if "new" == vmtype then
            vmfile = vmApp.moveMsg(vmbox, vmfile)
            vmApp.updateVM(vmbox)
        end
        if not vmfile or "" == vmfile then
            return false, "No voicemail file to playback"
        end
        app.Playback(vm_path .. vmfile)
    end

    local digits_num = 3
    while digits_num <= 10 do
        local pattern = string.format("_%s[.][+0-9].", string.rep("X", digits_num))
        extensions["pbxservices"][pattern] = function (c, e)
            local identity, num = string.match(e, "^(%d+)%.(.*)$")
            channel["__FROM_IDENTITY"]:set(identity or "")
            return app.Goto(c, num, 1)
        end
        digits_num = digits_num + 1
    end
end
