--- Dial by name / Directory service implementation
-- 
-- @file	init.lua 
-- @author  Gennadiy Ganchev
-- @created 2017.12.29

package.path = package.path..";/var/lib/callweaver/dialplan/stdService/Directory/?.lua"

-- @module Directory
local Directory = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local searchFunction = require"lookUp".byName
local searchFilterCallbacks = {
	 ["f"] = searchFunction
	,["l"] = searchFunction
	,["b"] = searchFunction
	,["a"] = searchFunction --AI: Directory Virtual Assistant
	,["i"] = searchFunction -- IVR via ASR with Directory (WMS-5289)
}

local function __validateOptions(exten, opts)
	local rv   = nil
	local opts = "table" == type(opts) and opts or {}

	-- checking filter value
	if "a" == exten or "i" == exten then
		-- check license via tts
		if require"textToSpeech".available() then
			opts.filter = exten
			if tonumber(channel["ASR_IVR_ATTEMPTS"]:get()) then
				opts.retry = tonumber(channel["ASR_IVR_ATTEMPTS"]:get())
			else
				opts.retry  = 3
			end
		else
			print("TextToSpeech/SpeechToText are unavailable")
			logger:info("DialByName: TextToSpeech/SpeechToText are unavailable")
		end
	end
	opts.filter = opts.filter or "b"
	if "*wair*" == channel.EXTEN:get() and opts.filter ~= "a" then
		logger:info("DialByName: call from wair headset, but 'a' mode is unavalailable")
		return rv
	end
	if not searchFilterCallbacks[opts.filter] then
		return rv
	end
	if "i" == exten and opts.filter ~= exten then
		return rv
	end

	rv = {
		 filter = opts.filter
		,source = opts.source
		,limit  = opts.limit
		,retry  = opts.retry
	}

	return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
function Directory.run(exten, opts)
	local rv	= false

	-- validate options
	opts = __validateOptions(exten, opts)
	if not opts then
		return rv
	end

	-- initialize a directory processing object
	local processingObj, err = require"processing":new(opts, searchFilterCallbacks)
	if not processingObj then
		return false, err
	end

	logger:info("Start DialByName in '%s' mode", opts.filter)

	-- processing procedure
	local usersList = {}
	local state	    = "welcome"
	repeat
		if "welcome" == state then
			usersList = {}
			state     = processingObj:init(usersList)
		elseif "choice" == state then
			state = processingObj:choice(usersList)
		end
	until nil == state or check_hangup()

	if check_hangup() then
		-- make hagup by caller
		return true
	elseif usersList.dialExt then
		-- execute dial
		local d_dialplan = "internalcalls"
		local d_exten    = ""
		local d_priority = 1
		if "ei" == string.sub(usersList.dialExt, 1, 2) then
			d_exten = string.sub(usersList.dialExt, 3)
			if channel["ASR_IVR_DIALPLAN"]:get() and "" ~= channel["ASR_IVR_DIALPLAN"]:get() then
				d_dialplan = channel["ASR_IVR_DIALPLAN"]:get()
			else
				d_dialplan = channel["CONTEXT"]:get()
			end
		else
			d_dialplan = "internalcalls"
			d_exten = usersList.dialExt
		end
		logger:info("DialByName: call to %s,%s,%s", d_dialplan, d_exten, d_priority)
		smartApp("Goto", d_dialplan, d_exten, d_priority)
		rv = true
	end

	return rv
end

return Directory
