--- Artificial Intelligence directory assistant playback instance
-- 
-- @file    playbackAI.lua 
-- @author  Gennadiy Ganchev
-- @created 2017/12/29

--- @module playbackAI
local playbackAI = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local playback = {
		mOpts = {
			 retry     = 1
			,max_chars = 1
			,timeout   = 20
		}
	}

--- Playback greeting message, record an answer and recognize it
function playback:welcome(_, filter, retryLimit)
	
	if channel["ASR_IVR_PROMPT"]:get() and "" ~= channel["ASR_IVR_PROMPT"]:get() then
		ivr:play({tts = channel["ASR_IVR_PROMPT"]:get()})
	elseif channel["DIALBYNAME_PROMPT"]:get() and "" ~= channel["DIALBYNAME_PROMPT"]:get() then
		ivr:play({tts = channel["DIALBYNAME_PROMPT"]:get()})
	else
		 ivr:play({tts = self.tr:_t("dialbyname.welcome")})
	end
	local params = {
			 language	   = channel.CHANNEL("Language"):get()
			,["phrases[]"] = self.dictionary
		}
	local rv = self.stt.recognize(self.stt.record(6), params)
	if not rv then
		ivr:play({tts = self.tr:_t("dialbyname.ididnotunderstand")})
	end

	return rv
end

--- Ask user for a choice
function playback:selectAction(last)
  if last then
    ivr:play({tts = self.tr:_t("dialbyname.confirmation")})
  else
    ivr:play({tts = self.tr:_t("dialbyname.choice")})
  end
	local params = {
			 language      = channel.CHANNEL("Language"):get()
			,["phrases[]"] = {self.tr:_t("dialbyname.cancel"), self.tr:_t("dialbyname.next")}
		}
	local rv = self.stt.recognize(self.stt.record(3), params)
	if self.tr:_t("dialbyname.cancel"):lower() == rv then
		rv = "cancel"
	elseif self.tr:_t("dialbyname.next"):lower() == rv then
		rv = "next"
	else
		rv = "call"
	end
	return rv
end

--- Playback "There are no more compatible entries"
function playback:noNextEntry()
	local phrase = self.tr:_t("dialbyname.nomoreentries")
	return ivr:play({tts = phrase})
end

--- Playback "I am sorry I could not find anybody called"
function playback:notFoundEntry(entry)
	ivr:play({tts = string.format("%s: %s", self.tr:_t("dialbyname.notfound"), entry)}, {tts_fixup = true})
end

--- Playback found user data.
function playback:userData(data)
	if "table" ~= type(data) then
		return false
	end
	-- playback Name and Surname
	if data.name then
		ivr:tell({tts = string.format("%s: %s", self.tr:_t("dialbyname.calling"), data.name)}, {tts_fixup = true})
	end

	return true
end

--- Playback "Too many matches for <recognition result>"
function playback:tooManyMatches(searchVal)
	local phrase = string.format("%s: %s", self.tr:_t("dialbyname.toomanymatches"), searchVal)
	return ivr:play({tts = phrase})
end


function playbackAI:new(sFilter)
	sFilter = sFilter or "a"
	local stt = require"speechToText"
	if not stt then
		return false, "Service not available"
	end
	local tr = require"common.translation".translation
	tr:set_language(string.lower(channel.CHANNEL("Language"):get() or "en"))
	local obj = {}
	setmetatable(obj, {__index = playback})
	obj.dictionary = require"lookUp".allUsers("LDAP", 500, sFilter)
	obj.stt        = stt
	obj.tr         = tr

	return obj
end

return playbackAI