-- Standard directory assistant playback instance
--
-- @file    playbackStd.lua
-- @author  Gennadiy Ganchev
-- @created 2017.12.29

--- @module playbackStd
local playbackStd = {}
----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local playback = {
		mOpts = {
			 retry	 = 1
			,max_chars = 1
			,timeout   = 20
		}
	}

local function __getOpts(args)
	local rv = {
		retry	 = playback.mOpts.retry
		,max_chars = playback.mOpts.max_chars
		,timeout   = playback.mOpts.timeout
	}

	if "table" == type(args) then
		for k,v in pairs(args) do
			rv[k] = v
		end
	end

	return rv
end

--- Playback greeting message and read user input
function playback:welcome(limit, filter)
	local limit = limit or 0	-- can be deprecated in current implementation
	local filterStr = "b" == filter and "name or surname" or "n" == filter and "name" or "s" == filter and "surname" or ""
	local phrase	= string.format("Welcome to the directory. Please enter part of %s for search, followed by pound.", filterStr)
	return ivr:read(phrase, __getOpts({max_chars = limit}))
end

--- Ask user for a choice
function playback:selectAction()
	local phrase = "If this is the person you are looking for press 1 otherwise press 0."
	return ivr:read(phrase, __getOpts())
end

--- Playback "There are no more compatible entries"
function playback:noNextEntry()
	local phrase = "There are no more compatible entries"
	return ivr:tell(phrase, __getOpts())
end

--- Playback "No entries match your search"
function playback:notFoundEntry()
	local phrase = "No entries match your search"
	return ivr:tell(phrase, __getOpts())
end

-- stub for search
function playback:tooManyMatches(searchVal)
end

--- Playback found user data.
function playback:userData(data)
	if "table" ~= type(data) then
		return false
	end

	repeat
		if not data.ext then
			break
		end
		-- playback voice mail greet
		-- TODO: improvement to user playback via ivr tell object
		local greetMsg = "/var/spool/callweaver/voicemail/" .. data.ext .. "/greet"
		smartApp("Playback", greetMsg)
		if "SUCCESS" == channel.PLAYBACKSTATUS:get() then
			break
		end
		-- otherwise playback only user extension
		print("User does not have voicemail greet message. Playback only number ...")
		smartApp("SayAlpha", data.ext)
	until true

	return true
end

function playbackStd:new()
	local obj = {}
	setmetatable(obj, {__index = playback})
	
	return obj
end

return playbackStd