--- Service to login, common part
--
-- @file    .lua
-- @author  Gennadiy Ganchev
-- @created 2017/12/15

--- @module common
local common = {}
----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
function common.run(number,wpin)
	if (number and not string.find(number,"^%d+$")) or 
		(wpin and not string.find(wpin,"^%d+$")) then
		ivr:tell("The value you entered is not valid. Hangup")
		return smartApp("Hangup", hcause.CHANUNAVAIL)
	end
	
	if string.find(channel.SIP_HEADER("User-Agent"):get() or "", "W-AIR") then
		if "0"== number or string.find(channel.SIP_HEADER("To"):get() or "", "<sip:%*wair%*@") then
			return require"stdService.login.byCollaboration".run(number)
		end
	end

	-- getting number and password
	local userAuth = require"userAuthorize":new()
		userAuth:setMaxAttempts(3)
		userAuth:setTimeOut(7)
	if not userAuth:process(number, wpin) then
		return smartApp("Hangup", hcause.NORMAL)
	end

	local pbxutils = require"core.pbxutils"
	local cid      = gSpace["cidNum"]
	local attrs    = userAuth:getUserData()
	local number   = attrs.telephoneNumber
	local choice   = nil
	local options  = {	--	set options for tell intermediate message
		nobeep = true,
		retry = 1,
		timeout = 0.1,
		max_chars = 1,
		validate = function(choice)
			return pbxutils.in_table(tonumber(choice), {0,1})
		end;
		failure = function()
			return true
		end;
	}

	if not cid:find("unknown%d") and not cid:find("unknown$") and cid ~= number then
		choice = ivr:read("The phone is in use by " .. cid, options)
	end

	local src_pbx = pbxutils.getPbxDataBySerial()
	if not options.validate(choice) then
		if attrs.pbxowner and attrs.pbxowner ~= src_pbx.dn then
			choice = ivr:read("You are currently logged in on another pbx " .. attrs.pbxowner, options)
			-- WMS-9177
			local msg = "Can not move user %s to this pbx (%s), move users between pbxes via dialplan deprecated"
			local err = string.format(msg, number, src_pbx.pbxSerial)
			print(err)
			logger:warning(err)
			return smartApp("Hangup", hcause.FAILURE)
		end
		if not options.validate(choice) then
			-- prepare options for main message
			options.retry   = false
			options.nobeep  = false
			options.timeout = false
			options.failure = function()
				ivr:tell("The number you entered is incorrect, please try again")
			end
			choice = ivr:read("Press 1 to use this phone or 0 to logout", options)
		end
	end
	if not choice then
		ivr:tell("Good bye. Hangup")
		return smartApp("Hangup", hcause.NORMAL)
	end
	return require"stdService.login.setUser".run(attrs,choice)
end

return common
