--- Set user to device
--
-- @file    setUser.lua
-- @author  Gennadiy Ganchev
-- @created 2017/12/12

--- @module setUser
local setUser = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
function setUser.run(dst,choice)
	if not dst or "" == dst or not choice or "" == choice then
		return false
	end
	local dstAttrs = nil
	local dn       = nil
	if "table" == type(dst) then
		dstAttrs = dst
		dn       = dst.DN
	else
		dstAttrs, dn = get_obj_attrs(dst,nil,nil,true)
	end
	local pbxutils                  = require"core.pbxutils"
	local cid                       = gSpace["cidNum"] or ""
	local useragent                 = channel.SIP_HEADER("User-Agent"):get() or ""
	local header                    = channel.SIP_HEADER("From"):get() or ""
	local sipuri                    = channel.SIPURI:get() or ""
	local cidip                     = channel["X-SRC-ADDRESS"]:get() or ""
	local removeContactShellEscaped = ""
	local removeContact             = ""
	local number                    = dstAttrs.telephoneNumber
	local handset                   = ""
	if sipuri ~= "" then
		removeContact = "'" .. cid .. "@wildix' '" .. sipuri .. "'"
		removeContactShellEscaped = cid .. "@wildix_" .. sipuri
	end
	local cidname = header:match('"(.*)"%s*\<')
	local cidnameparam = not cidname and "" or 'callername='.. cidname
	local useragentnospaces = useragent:gsub(" ","_")

	-- Check wildix device & get mac address 
	-- Wildix WP480 7.50.10.3 001565113aca					 - WP
	-- Wildix W-AIR 02.00.00.05 9c7514340002 11:88:70:03:74' - W-AIR
	-- Wildix W01FXSR2 3.7.1404240 0001A8094A28				 - W01FXSR2
	local model, _, mac = string.match(useragent, "Wildix%s(.-)%s(.-)%s([0-9A-Fa-f]*)")
	if not model or model == "" then
		ivr:tell("Good bye. Hangup")
		print("Model not found")
		return smartApp("Hangup", hcause.NORMAL)
	end

	if model:find("W-AIR") and model:find("/") then
		model, handset = string.match(model, "(.*)/(.*)")
		logger:debug("Handset model: %s (%s)", handset, model)
	end

	if "W-AIR-SB" == model or "W-AIR-SP" == model or "W-AIR-SBR2" == model then
		model = "W-AIR"
	end

	if model ~= "W-AIR" then
		if not mac or mac == "" then
			-- get mac address by username from database (only first record limitation)
			local devconnection = require"core.wsqlite":new("devdb")
			local sql = [[ SELECT mac FROM autoprov WHERE param=%q AND mac IN (SELECT mac FROM autoprov WHERE param=%q AND value=%q) AND value LIKE '%%"model":"%s"%%' ]]
			local cursor = devconnection:execute(sql, "device", "username", cid, model)
            if cursor then
                mac = cursor:fetch()
                cursor:close()
            end
            devconnection:close()
			if not mac or mac == "" then
				ivr:tell("Good bye. Hangup")
                print("MAC not found")
				return smartApp("Hangup", hcause.NORMAL)
			end
		end
		mac = mac:lower()
	end

    local util = require"utils"
	local sip_password
	if choice == "1" and number and dn then
		local src_pbx = pbxutils.getPbxDataBySerial()
		if dstAttrs.pbxowner and dstAttrs.pbxowner ~= src_pbx.dn then
			-- WMS-9177
			local msg = "Can not move user %s to this pbx (%s), move users between pbxes via dialplan deprecated"
			local err = string.format(msg, number, src_pbx.pbxSerial)
			print(err)
			logger:warning(err)
			return smartApp("Hangup", hcause.FAILURE)
		end
		
		print("Use this phone")
		sip_password = dstAttrs["wuserPasswordSip"]
		if not sip_password or "" == sip_password then
			local err = string.format("Sip password is not found for %s", number)
			print(err)
			logger:debug(err)
			return smartApp("Hangup", hcause.FAILURE)
		end
	elseif choice == "0" then
		print("Remove this phone")
		number = util:get_free_unknown_account()
		sip_password = "1234"
	else
		return false
	end
	
	if model == "W-AIR" then
		print("Call to service from W-AIR")
		local wildixair = require"modules.wildixair"
		local options = {
			["username"] = number,
			["callername"] = cidname,
			["cid"] = cid,
			["userid"] = dstAttrs.uid,
			["useragent"] = useragent:gsub(" ","_"),
			["removecommand"] = "kamctlwrapper "..removeContact.." &"
		}
		wildixair.setuser(cidip,'80',options)
		print("Sync via sip notify", cid)
	else
		local is_fxs = useragent:find("FXS")
		local is_wp2 = useragent:find("WP%w+R[234]")
		if not require"common.helper".isLocalIP(cidip) or is_fxs or is_wp2 then -- phone from nat or device need check-sync
			if is_fxs then
				print("Call to service from FXS")
			elseif is_wp2 then
				print("Call to service from WPR2 family phone")
			else
				print("NAT detected")
			end
			if model:find("FXS") and not model:find("W01FXSR2") and cidname then
				local port = cidname:match(".*%-(%d+)")
				if port then
					util:registerFXS(mac, number, port)
				end
			else
				util:register(mac, number)
			end
			setpostaction {
				["name"] = "LOGIN",
				["type"] = "notify",
				["cid"] = cid,
				["removecommand"] = "kamctlwrapper "..removeContact -- command will execute to bg in notify_peer
			}
		else	-- phone in local network and not FXS device (used only by WP)
			print("NAT not detected")
			util:register(mac, number) -- update username in DB
			local b64password = require"mime".b64(sip_password)
			local nmapCommand = "nohup sudo wmsbind -o 'action=setuser,username=%s,password=%q,userid=%s,useragent=%s,%s,cid=%s,modifier=logout,removeContact=%s' %s %s >/dev/null 2>&1 &"
			setpostaction {
				["name"] = "LOGIN",
				["type"] = "nmap",
				["command"] = string.format(nmapCommand, number, b64password, dstAttrs.uid or "", useragentnospaces, cidnameparam, cid, removeContactShellEscaped, cidip, mac)
			}
		end

		if choice == "1" and (is_wp2 or string.find(model, "WP")) then
			if "yes" ~= channel.MULTILOGIN_SUPPORT:get() then
				local devconnection = require"core.wsqlite":new("devdb")
				local sql = [[ SELECT mac FROM autoprov WHERE param='device' AND mac IN (SELECT mac FROM autoprov WHERE param='username' AND value=%q) AND value LIKE '%%"model":"wp%%' ]]
				local sql_res = devconnection:rows(sql, number)
				local user_devices = {}
				if sql_res then
					for devmac in sql_res do
						table.insert(user_devices, devmac)
					end
				end
				devconnection:close()

				local loggedOut = false
				for _, devmac in pairs(user_devices) do
					if devmac ~= mac then
						loggedOut = true
						util:register(devmac)
						print("Device with mac '%s' has been logged out", devmac)
					end
				end
				if loggedOut then
					util:notify_peer(number)
					logger:info("Other devices of user '%s' have been logged out", number)
				else
					logger:info("User '%s' has logged in via feature code", number)
				end
			else
				logger:info("MULTILOGIN_SUPPORT is enabled")
			end
		end
	end
	channel.CHANNEL("hangup_handler_wipe"):set("hdlr,service,1")
	ivr:tell("Thank you")
	
	return smartApp("Hangup", hcause.NORMAL)
end

function setUser.sync(action)
	if action.type == "nmap" then
		os.execute(action.command)
		print(action.command)
		print("Sync via nmap")
	elseif action.type == "notify" then
		local util = require"utils"
		util:notify_peer(action.cid, 3, action.removecommand)
		if action.username then
			util:notify_peer(action.username, 3, action.removecommand)
		end
		print("Sync via sip notify", action.cid, action.username)
	end
end

return setUser
