--- Module for play user ring tone
--
-- @file    userRingtone.lua
-- @author  Gennadiy Ganchev
-- @created 2016/11/01

--- @module ringTone
local userRingtone = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Preparing 'Alert-Info' SIP header with ringtone definition
-- @param ringTone   The user ring tone feature
-- @param callType   The call type value of current call
-- @return formatted 'Alert-Info' header for setting it via predial
function userRingtone.play(c, e)
        local app = app
        local ringtone = (e or ""):match("^ringtone(.*)")
        app.NoCDR()
        local variable = channel.RINGTONEUSER:get()
        if not variable or "" == variable then
          -- Play ringtone from XML browser ringtones
            app.Playback("/var/lib/callweaver/sounds/00000/RingTones/"..ringtone)
        else
            -- Set alert-info header for collaboration custom ring settings
            local user, ringtone = string.match(variable, "^(%d+)|RINGTONE=.*\/([^|]+)")
            if not user or not ringtone then return false end
            ringtone = require"ringTone".getAlertHeader(ringtone, "internal")
            local dialTimeout = channel["GINTDIAL_TIMEOUT"]:get() or "65"
            local dialOpts = string.format("b(predial^userringtone^1(,%s))", ringtone)
            return smartApp("Dial", "SIP/"..user, dialTimeout, dialOpts)
        end
end

--- Pre-dial rules for user ringtone dial
function userRingtone.preDialRules()
    local ringer = channel.ARG2:get() or ""
    if ringer ~= "" then
        channel.SIP_ALL_ADDHEADER_RINGER:set(ringer)
    end
end

return userRingtone
