--- Wrapper for dialplan custom application
-- Loading dialplan without error and execution application via protected call
--
-- @file    customApplication.lua
-- @author  Vyacheslav Khudan
-- @created 2017/02/27

local logger = logger

--- @module customApplication
local customApplication = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
-- Originate via AMI
local function __originate()
    local ami = require"core.AMIClient":new()
    local o_channel = channel["ORIGINATE_CHANNEL"]:get() ~= "" and channel["ORIGINATE_CHANNEL"]:get() or nil
    local o_context = channel["ORIGINATE_CONTEXT"]:get() ~= "" and channel["ORIGINATE_CONTEXT"]:get() or nil
    local o_exten = channel["ORIGINATE_EXTEN"]:get() ~= "" and channel["ORIGINATE_EXTEN"]:get() or nil
    local o_priotity = channel["ORIGINATE_PRIORITY"]:get() ~= "" and channel["ORIGINATE_PRIORITY"]:get() or nil
    local o_app = channel["ORIGINATE_APP"]:get() ~= "" and channel["ORIGINATE_APP"]:get() or nil
    local o_data = channel["ORIGINATE_DATA"]:get() ~= "" and channel["ORIGINATE_DATA"]:get() or nil
    local o_clid = channel["ORIGINATE_CALLERID"]:get() ~= "" and channel["ORIGINATE_CALLERID"]:get() or nil
    local o_timeout = channel["ORIGINATE_TIMEOUT"]:get() ~= "" and tonumber(channel["ORIGINATE_TIMEOUT"]:get()) or nil
    if o_timeout then
        o_timeout = o_timeout * 1000
    end
    local o_variables = channel["ORIGINATE_VARIABLES"]:get() ~= "" and channel["ORIGINATE_VARIABLES"]:get() or nil
    local o_async = channel["ORIGINATE_ASYNC"]:get():lower() == "true" and true or false

    logger:debug({
        channel     = o_channel,
        context     = o_context,
        exten       = o_exten,
        priotity    = o_priotity,
        application = o_app,
        data        = o_data,
        callerid    = o_clid,
        timeout     = o_timeout,
        variables   = o_variables,
        async       = o_async
        })

    st, errors = ami:originate(o_channel, o_context, o_exten, o_priotity, o_app, o_data, o_clid, o_timeout, 
                               o_variables, o_async)
    if not st then
        print("Errors during originate: " .. tostring(errors))
        logger:error("Errors during originate: " .. tostring(errors))
    end
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local application = {}  -- meta class

--- Main procedure to make executing application
-- @param args  An custom application arguments 
-- @return boolean value as an operation result
function application:exec(args)
    repeat
        if not args then
            logger:warning("Custom application has empty arguments!")
            break
        end
        local appName, appArgs = args:match("^([^%(%s]+)%s*%((.*)%)")
        if not appName or not appArgs then
            logger:debug("Custom application arguments parsing error! Use format 'NoOp(Application data)': %s", args)
            break
        end
        logger:info("Executing custom app.%s(%s)", appName, appArgs)
        logger:info({customapp = {appName = appName, appArgs = appArgs}})

        if "originate" == appName:lower() then
            __originate()
            break
        end

        if appArgs:find("${.+}") then
            appArgs = channel.EVAL(appArgs):get() or ""
            logger:debug("Evaluated custom app.%s(%s)", appName, appArgs)
        end
        -- NOTE: user's syntax errors cause lua exceptions and ugly callweaver log with traces
        -- but do not use pcall() here due to callweaver lua magic
        return app[appName](appArgs)
    until true
end

return setmetatable(customApplication,  {
     __index = application
    ,__call  = application.exec
})
